# coding: utf-8
# Standard Python libraries
from collections import OrderedDict
from copy import deepcopy
import warnings

# http://www.numpy.org/
import numpy as np

# https://pandas.pydata.org/
import pandas as pd

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# atomman imports
import atomman.unitconvert as uc
from . import Atoms, Box, dvect, dmag, NeighborList
from ..lammps import normalize as lmp_normalize
from ..tools import indexstr, miller, ishexagonal, aslist
from .. import dump

class System(object):
    """
    A representation of an atomic system.  This combines Atoms with Box and
    adds methods and attributes involving both.
    """
    
    class _AtomsIndexer(object):
        """Internal class for setitem / getitem acting on atoms"""
        def __init__(self, host):
            self.__host = host
            
        def __getitem__(self, index):
            """Index getting of Atoms that operates on System."""
            host = self.__host
            return System(atoms=host.atoms[index], box=host.box, pbc=host.pbc,
                          symbols=host.symbols)
            
        def __setitem__(self, index, value):
            """Index setting of Atoms that operates on System."""
            host = self.__host
            if isinstance(value, Atoms):
                host.atoms[index] = value
            elif isinstance(value, System):
                try:
                    assert np.allclose(host.box.vects, value.box.vects)
                    assert np.allclose(host.box.origin, value.box.origin)
                except:
                    warnings.warn('Atom assignment between two Systems with different boxes', UserWarning)
                host.atoms[index] = value.atoms
            else:
                raise ValueError('Can only set using Atoms or System objects')
    
    def __init__(self, atoms=None, box=None, pbc=None, scale=False,
                 symbols=None, masses=None, model=None, safecopy=False):
        """
        Initialize a System by joining an am.Atoms and am.Box instance.
        
        Parameters
        ----------
        atoms : atomman.Atoms, optional
            The underlying Atoms object to build system around.
        box : atomman.Box, optional
            The underlying box object to build system around.
        pbc : tuple or list of bool, optional
            Indicates which of the dimensions related to the three box vectors
            are periodic.  Default value is (True, True, True).
        scale : bool, optional
            If True, atoms.pos will be scaled relative to the box.  Default
            value is False.
        symbols : tuple, optional
            A list of the element symbols for each atom atype.  If len(symbols)
            is less than natypes, then missing values will be set to None.
            Default sets list with all None values.
        masses : tuple, optional
            A list of the masses for each atom atype.  If len(symbols) is less
            than natypes, then missing values will be set to None.  Default
            sets list with all None values.
        model : str or DataModelDict, optional
            File path or content of a JSON/XML data model containing all
            system information.  Cannot be given with atoms, box or scale.
        safecopy : bool, optional
            Flag indicating if values are to be copied before setting.  For
            values given as objects, direct setting (False, default) may result
            in the System pointing to the original object.  Using safecopy=True
            deep copies the objects before setting to avoid this.  Note that
            safecopy=True may be considerably slower for large numbers of atoms
            and/or properties.
        
        Returns
        =======
        System
            The System object.
        """
        # Check for model
        if model is not None:
            try:
                assert atoms is None
                assert box is None
                assert scale is False
            except:
                raise ValueError('model cannot be given with atoms, box or scale parameters')
            
            # Load data model
            model = DM(model).find('atomic-system')

            # Extract values from model
            box = Box(model=model)
            atoms = Atoms(model=model)
            if pbc is None:
                pbc = model['periodic-boundary-condition']
            if symbols is None:
                symbols = tuple(model.aslist('atom-type-symbol'))
            if masses is None:
                masses = tuple(model.aslist('atom-type-mass'))

        # Interpret given/missing parameters
        else:
            # Set default atoms or deepcopy
            if atoms is None:
                atoms = Atoms()
            elif safecopy:
                atoms = deepcopy(atoms)

            # Set default box or deepcopy
            if box is None:
                box = Box()
            elif safecopy:
                box = deepcopy(box)
            
            # Set default pbc
            if pbc is None:
                pbc = (True, True, True)
            
            # Set default masses
            if masses is None:
                masses = []
            else:
                masses = aslist(masses)
            
            # Set default symbols
            if symbols is None:
                symbols = [None for i in range(len(masses))]
            else:
                symbols = aslist(symbols)
            
            # Check data types
            if not isinstance(atoms, Atoms):
                raise TypeError('Invalid atoms type')
            if not isinstance(box, Box):
                raise TypeError('Invalid box type')
            if not isinstance(scale, bool):
                raise TypeError('Invalid scale type')
        
        # Set properties
        self.__atoms = atoms
        self.__box = box
        self.pbc = pbc
        self.__transformation = np.identity(3)

        self.symbols = symbols
        self.masses = masses
        
        # Scale pos if needed
        if scale is True:
            self.atoms_prop('pos', value=self.atoms.pos, scale=True)
        
        # Scale model properties if needed
        if model is not None:
            for prop in model['atoms'].aslist('property'):
                if prop['data'].get('unit', None) == 'scaled':
                    self.atoms.view[prop['name']] = self.unscale(self.atoms.view[prop['name']]) 
        
        # Set atoms indexer
        self.__atoms_ix = System._AtomsIndexer(self)
    
    def __str__(self):
        """str : The string representation of a system."""
        return '\n'.join([str(self.box),
                          'natoms = ' + str(self.natoms),
                          'natypes = ' + str(self.natypes),
                          'symbols = ' + str(self.symbols),
                          'pbc = ' + str(self.pbc),
                          str(self.atoms)])
    
    def __len__(self):
        return self.natoms
    
    @property
    def atoms(self):
        """atomman.Atoms : underlying Atoms object."""
        return self.__atoms
    
    @property
    def natoms(self):
        """int : The number of atoms in the Atoms class."""
        return self.__atoms.natoms
    
    @property
    def atypes(self):
        """tuple : List of int atom types."""
        return tuple(range(1, self.natypes+1))
    
    @property
    def natypes(self):
        """int : The number of atom types/symbols."""
        
        # Check if more symbols than atypes
        try:
            nsymbols = len(self.symbols)
        except:
            nsymbols = 0
        if nsymbols > self.__atoms.natypes:
            return len(self.symbols)
        else:
            return self.__atoms.natypes
    
    @property
    def box(self):
        """atommman.Box : underlying Box object."""
        return self.__box
    
    @property
    def pbc(self):
        """list of bool : The periodic boundary condition settings."""
        return self.__pbc
        
    @pbc.setter
    def pbc(self, value):
        pbc = np.asarray(value, dtype=bool)
        assert pbc.shape == (3,), 'invalid pbc entry' 
        self.__pbc = pbc
    
    @property
    def atoms_ix(self):
        """Indexer for index slicing of Systems by per-atom properties."""
        return self.__atoms_ix

    @property
    def symbols(self):
        """tuple : The element model symbols associated with each atype."""
        
        # Fill in missing values
        if len(self.__symbols) < self.__atoms.natypes:
            self.symbols = self.__symbols
        return self.__symbols
    
    @symbols.setter
    def symbols(self, value):
        
        # Make value list if needed
        value = aslist(value)
        
        # Fill in missing values
        if len(value) < self.__atoms.natypes:
            newvalue = [None for x in range(self.__atoms.natypes)]
            for i in range(len(value)):
                newvalue[i] = value[i]
            value = newvalue

        self.__symbols = tuple(value)

    @property
    def masses(self):
        """tuple : The masses associated with each atype if given."""
        
        # Fill in missing values
        if len(self.__masses) < self.natypes:
            self.masses = self.__masses
        return self.__masses
    
    @masses.setter
    def masses(self, value):
        
        # Make value list if needed
        value = aslist(value)
        
        # Convert non-None values to float
        for i in range(len(value)):
            if value[i] is not None:
                value[i] = float(value[i])

        # Fill in missing values
        if len(value) < self.natypes:
            newvalue = [None for x in range(self.natypes)]
            for i in range(len(value)):
                newvalue[i] = value[i]
            value = newvalue
        elif len(value) > self.natypes:
            raise ValueError('More masses than atom types given. Either change atype values or symbols first.')

        self.__masses = tuple(value)
    
    @property
    def composition(self):
        """
        The system's reduced and sorted symbols composition.
        
        Returns
        -------
        str or None
            The system's reduced and sorted symbols composition.
            Will return None if symbols are missing
        """
        
        # Build total count of each symbol
        sym_dict = {}
        for i in range(self.natypes):
            count = np.sum(self.atoms.atype == i+1)
            if count > 0:
                symbol = self.symbols[i]
                
                if symbol is None:
                    return None
                
                if symbol in sym_dict:
                    sym_dict[symbol] += count
                else:
                    sym_dict[symbol] = count

        # Find greatest common denominator
        gcd = np.gcd.reduce(list(sym_dict.values())) # pylint: disable=no-member
        
        # Sort symbols and reduce counts
        composition =''
        for symbol in sorted(sym_dict):
            count = sym_dict[symbol] // gcd
            if sym_dict[symbol] > 0:
                composition += symbol
                if count != 1:
                    composition += str(count)
        
        return composition
    
    def atoms_prop(self, key=None, index=None, value=None, a_id=None, scale=False):
        """
        Extends Atoms.prop() by adding a scale argument.
        
        Parameters
        ----------
        key : str, optional
            Per-atom property name.
        index : int, list, slice, optional
            Index of atoms.
        value : any, optional
            Property values to assign.
        a_id : int, optional
            Integer atom index.  Left in for backwards compatibility.
        scale : bool, optional
            Flag indicating if values should be scaled/unscaled. If True:
            - Values being retrieved are scaled from absolute Cartesian to box relative vectors.
            - Values being set are unscaled from box relative to absolute Cartesian vectors.
            Default value is False.
        """
        
        # Check that scale is bool
        if not isinstance(scale, bool):
            raise TypeError('Invalid scale type')
        
        # Call atoms.prop() for scale is False
        if scale is False:
            if value is None:
                return self.atoms.prop(key=key, index=index, a_id=a_id)
            else:
                self.atoms.prop(key=key, index=index, value=value, a_id=a_id)
        
        # Handle property scaling if scale is True
        else:
            # Handle a_id
            if a_id is not None:
                if index is not None:
                    raise ValueError('a_id and index cannot both be given')
                index = a_id
            
            # Get values if value is None
            if value is None:
                # If no key, return copy of atoms (all or a slice) with scaled pos
                if key is None:
                    if index is None:
                        newatoms = deepcopy(self.atoms)
                    else:
                        newatoms = deepcopy(self.atoms[index])
                    newatoms.pos = self.scale(newatoms.pos)
                    return newatoms
                
                # If key is given, return scaled property values
                else:
                    if index is None:
                        value = self.atoms.view[key]
                    else:
                        value = self.atoms.view[key][index]
                    return self.scale(value)
            
            # Set values if value is given
            else:
                # If no key, unscale pos of Atoms value and set to atoms (or a slice)
                if key is None:
                    if not isinstance(value, Atoms):
                        raise TypeError('If key is None, value must be instance of atomman.Atoms')
                    
                    value.pos = self.unscale(value.pos)
                    if index is None:
                        self.atoms[:] = value
                    else:
                        self.atoms[index] = value
                
                # If key is given, unscale value and set to property
                else:
                    value = self.unscale(value)
                    if index is None:
                        self.atoms.view[key] = value
                    else:
                        self.atoms.view[key][index] = value
    
    def atoms_df(self, scale=False):
        """
        Extends Atoms.df() by adding a scale argument.
        
        Parameters
        ----------
        scale : bool or list, optional
            Indicates if/which per-atom properties are to be scaled to box
            relative values.  If False (default), no properties will be scaled.
            If True, pos will be scaled.  Multiple properties can be scaled by
            providing a list of the property names to scale.
            
        Returns
        -------
        pandas.DataFrame
            Tabulated DataFrame version of the atomic data.
        """
        # Handle scale values
        if scale is True:
            scale = ['pos']
        elif scale is False:
            scale = []
        elif not isinstance(scale, list):
            scale = [scale]
        
        # Initialize new dictionary of values
        values = OrderedDict()
        for key in self.atoms.view.keys():
            
            value = self.atoms.view[key]
            if key in scale:
                value = self.scale(value)
            
            # Flatten multidimensional arrays
            for index, istr in indexstr(self.atoms.view[key].shape[1:]):
                newkey = key + istr
                
                # Copy values over
                if index == ():
                    values[newkey] = value
                else:
                    values[newkey] = value[(Ellipsis, ) + index]
        
        # Return DataFrame
        return pd.DataFrame(values)
    
    def atoms_extend(self, value, scale=False, symbols=None, safecopy=False):
        """
        Extends Atoms.extend() to the System level by adding scale and symbols
        arguments.
        
        Parameters
        ----------
        value : atomman.Atoms or int
            An int value will result in the atoms object being extended by
            that number of atoms, with all per-atom properties having default
            values (atype = 1, everything else = 0).  For an Atoms value, the
            current atoms list will be extended by the correct number of atoms
            and all per-atom properties in value will be copied over.  Any
            properties defined in one Atoms object and not the other will be
            set to default values.
        scale : bool, optional
            Flag indicating if position values in a supplied Atoms value are to
            be taken as absolute Cartesian (False, default) or in scaled box
            relative units (True).
        symbols : tuple, list or None, optional
            Allows for the system's symbols list to be updated.  If not given,
            will use the current object's symbols.
        safecopy : bool, optional
            Flag indicating if values are to be copied before setting.  If 
            False (default), underlying objects may be shared between the new
            system and the current system and input parameters.  If True, atoms
            and box will be deepcopied before setting.
            Note that safecopy=True may be considerably slower for large
            numbers of atoms and/or properties.
        
        Returns
        -------
        atomman.System
            A new System object with Atoms extended to contain all atoms and
            properties of the current object plus the additional atoms.  The
            cuurent System object's box and pbc (and symbols if not specified)
            will be copied over.
        """
        # Copy value if safecopy
        if safecopy:
            value = deepcopy(value)
        
        # scale only makes sense for Atoms values
        if scale is True and not isinstance(value, Atoms):
            raise ValueError('scale can only be True for Atoms values')        
        
        # Handle symbols parameter
        if symbols is None:
            symbols = self.symbols
        
        # Handle box
        if safecopy:
            box = deepcopy(self.box)
        else:
            box = self.box
        
        # Call atoms.extend to generate new atoms
        atoms = self.atoms.extend(value)
        
        # Unscale pos from Atoms value if needed
        if scale:
            atoms.pos[value.natoms:] = self.unscale(value.pos)

        # Generate and return new System
        return System(atoms=atoms, box=box, pbc=self.pbc, symbols=symbols)
    
    def box_set(self, **kwargs):
        """
        Extends box.set() with a scale argument.
        
        Parameters
        ----------
        scale : bool
            If True, the scaled (box-relative) positions remain unchanged.
            Default value is False (absolute positions unchanged).
        other kwargs : any
            Any other kwargs allowed by box.set().
        """
        
        # Pop scale
        scale = kwargs.pop('scale', False)
        if not isinstance(scale, bool):
            raise TypeError('Invalid scale type')
        
        # Hold scaled positions constant
        if scale is True:
            spos = self.atoms_prop('pos', scale=True)
            self.box.set(**kwargs)
            self.atoms_prop('pos', value=spos, scale=True)
        
        # Call box.set without scaling
        else:
            self.box.set(**kwargs)
    
    def scale(self, value):
        """
        Scales 3D vectors from absolute Cartesian coordinates to relative box
        coordinates.
        
        Parameters
        ----------
        value : numpy.ndarray
            Values to scale.
        """
        
        # Retrieve parameters
        value = np.asarray(value, dtype=float)
        vects = self.box.vects
        inverse = np.linalg.inv(vects)
        origin = self.box.origin
        
        # Convert
        return (value - origin).dot(inverse)
          
    def unscale(self, value):
        """
        Unscales 3D vectors from relative box coordinates to absolute
        Cartesian coordinates.
        
        Parameters
        ----------
        value : numpy.ndarray
            Values to unscale.
        """
        
        # Retrieve parameters
        value = np.asarray(value, dtype=float)
        vects = self.box.vects
        origin = self.box.origin
        
        # Convert
        return value.dot(vects) + origin
        
    def wrap(self, return_imageflags=False):
        """
        Wrap atoms around periodic boundaries and extend non-periodic
        boundaries such that all atoms are within the box.

        Parameters
        ----------
        return_imageflags : bool, optional
            If True, an array of which image the atom was originally in is
            returned.

        Returns
        -------
        numpy.NDarray of int
            The imageflags array
        """
        
        # mins and maxs are box dimensions relative to box vectors, i.e 0 to 1
        mins = np.array([0.0, 0.0, 0.0])
        maxs = np.array([1.0, 1.0, 1.0])
        
        # Retrieve scaled pos
        spos = self.atoms_prop('pos', scale=True)

        # Initialize wrapflags
        imageflags = np.zeros_like(spos, dtype=np.int)
        
        # Loop over three pbc directions
        for i in range(3):
            
            # Count wraps across periodic boundaries
            if self.pbc[i]:
                imageflags[:, i] = np.floor(spos[:, i]) # pylint: disable=unsupported-assignment-operation
            
            # Shift min and max to encompass atoms across non-periodic bounds
            else:
                min = spos[:, i].min()
                max = spos[:, i].max()
                if min <= mins[i]: 
                    mins[i] = min - 0.001
                if max >= maxs[i]: 
                    maxs[i] = max + 0.001
            
        # Wrap atoms across periodic boundaries
        spos -= imageflags
        
        # Unscale spos and save to pos
        self.atoms_prop('pos', value=spos, scale=True)
        
        # Modify box vectors and origin by new min and max
        origin = self.box.origin + mins.dot(self.box.vects) 
        avect = self.box.avect * (maxs[0] - mins[0])
        bvect = self.box.bvect * (maxs[1] - mins[1])
        cvect = self.box.cvect * (maxs[2] - mins[2])
        self.box_set(avect=avect, bvect=bvect, cvect=cvect, origin=origin)

        if return_imageflags:
            return imageflags
    
    def dvect(self, pos_0, pos_1):
        """
        Computes the shortest vector between pos_0 and pos_1 using box 
        dimensions and accounting for periodic boundaries.
        
        Parameters
        ----------
        pos_0 : numpy.ndarray or index
            Absolute Cartesian vector position(s) to use as reference point(s).
            If the value can be used as an index, then self.atoms.pos[pos_0]
            is taken.
        pos_1 : numpy.ndarray or index
            Absolute Cartesian vector position(s) to find relative to pos_0.
            If the value can be used as an index, then self.atoms.pos[pos_1]
            is taken.

        Returns
        -------
        numpy.ndarray
            The shortest vectors from each pos_0 to pos_1 positions.
        """
        # Test if pos_0 and pos_1 can be used as numpy array indices
        try:
            pos_0 = self.atoms.pos[pos_0]
        except:
            pos_0 = np.asarray(pos_0)
        try:
            pos_1 = self.atoms.pos[pos_1]
        except:
            pos_1 = np.asarray(pos_1)
        
        # Call dvect using self's box and pbc
        vects = dvect(pos_0, pos_1, self.box, self.pbc)
        if len(vects) == 1:
            return vects[0]
        else:
            return vects

    def dmag(self, pos_0, pos_1):
        """
        Computes the shortest distance between pos_0 and pos_1 using box 
        dimensions and accounting for periodic boundaries.
        
        Parameters
        ----------
        pos_0 : numpy.ndarray or index
            Absolute Cartesian vector position(s) to use as reference point(s).
            If the value can be used as an index, then self.atoms.pos[pos_0]
            is taken.
        pos_1 : numpy.ndarray or index
            Absolute Cartesian vector position(s) to find relative to pos_0.
            If the value can be used as an index, then self.atoms.pos[pos_1]
            is taken.
        
        Returns
        -------
        numpy.ndarray
            The shortest vector magnitude from each pos_0 to pos_1 positions.
        """
        # Test if pos_0 and pos_1 can be used as numpy array indices
        try:
            pos_0 = self.atoms.pos[pos_0]
        except:
            pos_0 = np.asarray(pos_0)
        try:
            pos_1 = self.atoms.pos[pos_1]
        except:
            pos_1 = np.asarray(pos_1)
        
        # Call dvect using self's box and pbc
        vects = dmag(pos_0, pos_1, self.box, self.pbc)
        if len(vects) == 1:
            return vects[0]
        else:
            return vects
    
    def neighborlist(self, **kwargs):
        """
        Builds a neighbor list for the system.  The resulting NeighborList
        object is saved to the object as attribute 'neighbors'.
        
        Parameters
        ----------
        cutoff : float, optional
            Radial cutoff distance for identifying neighbors.  Must be given if
            model is not given.
        model : str or file-like object, optional
            Gives the file path or content to load.  If given, no other
            parameters are allowed.
        initialsize : int, optional
            The number of neighbor positions to initially assign to each atom.
            Default value is 20.
        deltasize : int, optional
            Specifies the number of extra neighbor positions to allow each atom
            when the number of neighbors exceeds the underlying array size.
            Default value is 10.
            
        Returns
        -------
        atomman.NeighborList
            The compiled list of neighbors.
        """
        if 'system' in kwargs:
            raise KeyError("Parameter 'system' not allowed")
        else:
            kwargs['system'] = self
        return NeighborList(**kwargs)
    
    def r0(self, neighbors=None):
        """
        Identifies the shortest interatomic spacing between atoms in the
        system by comparing the shortest periodic box vector with the
        smallest dmags of all neighbor atoms.
        
        Parameters
        ----------
        neighbors : NeighborList, optional
            A pre-computed NeighborList for the system.  If not given, a new
            NeighborList will be computed using a cutoff distance based on the
            smallest dmag between atom 0 and the rest of the system's atoms.
        
        Returns
        -------
        float
            The shortest interatomic spacing identified.
        """
        
        # Find shortest periodic lattice parameter
        try:
            box_r0 = np.linalg.norm(self.box.vects, axis=1)[self.pbc].min()
        except:
            box_r0 = None
        
        # Find shortest interatomic vector
        if self.natoms > 1:
            
            # Compute neighbors if needed
            if neighbors is None:
                
                # Use r0 for atom 0 as cutoff
                cutoff = self.dmag(0, range(1, self.natoms)).min() * 1.01
                
                # Identify all neighbors
                neighbors = self.neighborlist(cutoff=cutoff)
            
            # Find smallest r0 across all neighbor sets
            atom_r0 = np.inf
            for i in range(self.natoms):
                j = neighbors[i]
                if len(j) > 0:
                    new_r0 = self.dmag(i, j).min()
                    if new_r0 < atom_r0:
                        atom_r0 = new_r0

            if atom_r0 == np.inf:
                atom_r0 = None
        else:
            atom_r0 = None

        if box_r0 is not None and atom_r0 is not None:
            if atom_r0 < box_r0:
                return atom_r0
            else:
                return box_r0
        elif box_r0 is not None:
            return box_r0
        elif atom_r0 is not None:
            return atom_r0
        else:
            raise ValueError('No atoms to compare or periodic boundaries found!')
    
    def supersize(self, a_size, b_size, c_size):
        """
        Creates a larger system from a given system by replicating it along the
        system's box vectors.
        
        The multiplier values \\*_size are taken to be integer tuples (m, n) where
        m <= 0 and n >= 0.  The system multiplication works such that if n = -m,
        then the seed system's origin will be at the center of the new system.
        If only one integer is given, then it is assigned to m or n depending on
        its sign, and the other value is taken to be 0.
        
        Parameters
        ----------
        a_size : int or tuple of int
            Single int or two integers specifying replication along the avect
            direction.
        b_size -- int or tuple of int
            Single int or two integers specifying replication along the bvect
            direction.
        c_size -- int or tuple of int
            Single int or two integers specifying replication along the cvect
            direction.
        
        Returns
        -------
        atomman.System
            A new system created by replicating the given seed system according to
            the \\*_size parameters.
          
        """
        # Extract parameters
        sizes = [a_size, b_size, c_size]
        mults = np.array([0, 0, 0], dtype=int)
        vects = self.box.vects
        origin = self.box.origin
        spos = self.atoms_prop('pos', scale=True)
        
        # Check the *_size values
        for i in range(3):
            
            # Change single int to tuple of two int
            if isinstance(sizes[i], (int, np.integer)):
                if sizes[i] > 0:
                    sizes[i] = (0, sizes[i])
                elif sizes[i] < 0:
                    sizes[i] = (sizes[i], 0)
            
            elif isinstance(sizes[i], tuple):
                try:
                    assert len(sizes[i]) == 2, str(len(sizes[i]))
                    assert isinstance(sizes[i][0], (int, np.integer)), str(sizes[i][0])
                    assert sizes[i][0] <= 0, str(sizes[i][0])
                    assert isinstance(sizes[i][1], (int, np.integer)), str(sizes[i][1])
                    assert sizes[i][1] >= 0, str(sizes[i][1])
                except:
                    raise TypeError('Invalid system multipliers')
            else:
                raise TypeError('Invalid system multipliers')
            
            # Calculate full multipliers 
            mults[i] = sizes[i][1] - sizes[i][0]
            if mults[i] == 0:
                raise ValueError('Cannot multiply system dimension by zero')
                
            # Scale box and first set of positions accordingly
            spos[:,i] /= mults[i]
            origin += vects[i] * sizes[i][0]
            vects[i] *= mults[i]
            
        # Initialize new Box and Atoms
        box = Box(vects=vects, origin=origin)
        natoms = self.natoms * mults[0] * mults[1] * mults[2]
        atoms = Atoms(natoms=natoms)
        
        # Copy over all property values (except pos) using numpy broadcasting
        for key in self.atoms_prop():
            if key == 'pos':
                continue
        
            # Get old array
            old = self.atoms.view[key]
            
            # Create new array and broadcast old to it
            new = np.empty((mults[0] * mults[1] * mults[2],) + old.shape, dtype = old.dtype)
            new[:] = old
            
            # Reshape new and save to atoms
            new_shape = new.shape
            new_shape = (new_shape[0] * new_shape[1],) + new_shape[2:] # pylint: disable=unsubscriptable-object
            atoms.view[key] = np.array(new.reshape(new_shape))
        
        # Expand spos using broadcasting
        new_spos = np.empty((mults[0] * mults[1] * mults[2],) + spos.shape)
        new_spos[:] = spos
        
        # Reshape spos
        new_shape = new_spos.shape
        new_shape = (new_shape[0]*new_shape[1],) + new_shape[2:] # pylint: disable=unsubscriptable-object
        new_spos = new_spos.reshape(new_shape)
        
        # Use broadcasting to create arrays to add to spos
        test = np.empty(mults[0] * self.natoms)
        test.shape = (self.natoms, mults[0])
        test[:] = np.arange(mults[0])
        x = test.T.flatten()

        test = np.empty(mults[1] * len(x))
        test.shape = (len(x), mults[1])
        test[:] = np.arange(mults[1])
        y = test.T.flatten()
        test.shape = (mults[1], len(x))
        test[:] = x
        x = test.flatten()

        test = np.empty(mults[2] * len(x))
        test.shape = (len(x), mults[2])
        test[:] = np.arange(mults[2])
        z = test.T.flatten()
        test.shape = (mults[2], len(x))
        test[:] = x
        x = test.flatten()
        test[:] = y
        y = test.flatten()
        
        # xyz is displacement values to add to spos
        xyz = (np.hstack((x[:, np.newaxis], y[:, np.newaxis], z[:, np.newaxis])) 
             * np.array([1 / mults[0], 1 / mults[1], 1 / mults[2]]))
        
        # Save pos values, return new System
        atoms.view['pos'] = new_spos + xyz
        
        return System(box=box, atoms=atoms, scale=True, symbols=self.symbols)
    
    def rotate(self, uvws, tol=None, return_transform=False):
        """
        Transforms a System representing a periodic crystal cell from a standard
        orientation to a specified orientation. Note: if hexagonal indices are
        given, the vectors will be reduced to the smallest uvw integer
        representation.
        
        Parameters
        ----------
        uvws : numpy.ndarray
            A (3, 3) array of the Miller crystal vectors or a (3, 4) array of
            Miller-Bravais hexagonal crystal vectors to use in transforming the
            system.  Values must be integers.
        tol : list or float, optional
            Tolerance parameter used in determining which atoms are inside the
            box.  Multiple values can be given as the identification may
            occasionally fail for a given ucell and tol.  Default behavior will
            try tol values ranging from 1e-4 to 1e-8.
        return_transform : bool, optional
            Indicates if the transformation matrix associated with the
            rotation is returned.  Default value is False.
        
        Returns
        -------
        atomman.System
            A new fully periodic system rotated and transformed according to the
            uvws crystal vectors.
        transform : np.ndarray
            The transformation matrix associated with the rotation.
            Returned if return_transform is True.
        """
        
        if tol is None:
            tol = [1e-4, 1e-5, 1e-6, 1e-7]
        else:
            tol = aslist(tol)
        
        uvws = np.asarray(uvws)
        
        # Convert uvws from Miller-Bravais to Miller indices if needed
        if uvws.shape == (3, 4):
            if ishexagonal(self.box):
                uvws = miller.vector4to3(uvws)
            else:
                raise ValueError('hexagonal indices only work on hexagonal systems')
        
        # Check uvws shape and values
        if uvws.shape != (3, 3):
            raise ValueError('Invalid uvws crystal indices shape')

        int_uvws = np.asarray(np.rint(uvws), dtype='int64')
        if np.allclose(uvws, int_uvws):
            uvws = int_uvws
        else:
            raise ValueError('Rotation uvws must be integer values')
        
        # Get natoms and volume of system
        natoms = self.natoms
        volume = self.box.volume
        
        # Convert uvws to Cartesian units and compute new volume and natoms
        newvects = miller.vector_crystal_to_cartesian(uvws, box=self.box)
        newvolume = np.abs(newvects[0].dot(np.cross(newvects[1], newvects[2])))
        newnatoms = int(round(newvolume / volume) * natoms)
        
        # Check new values
        if newnatoms == 0:
            raise ValueError('New box has no atoms/volume: vectors are parallel or planar')
        
        # Identify box corners of new system wrt uvws
        corners = np.empty((8,3), dtype='int64')
        corners[0] = np.zeros(3)
        corners[1] = uvws[0]
        corners[2] = uvws[1]
        corners[3] = uvws[2]
        corners[4] = uvws[0] + uvws[1]
        corners[5] = uvws[0] + uvws[2]
        corners[6] = uvws[1] + uvws[2]
        corners[7] = uvws[0] + uvws[1] + uvws[2]
        
        # Create a supercell of system that contains all box corners
        a_mults = (corners[:,0].min()-1, corners[:,0].max()+1)
        b_mults = (corners[:,1].min()-1, corners[:,1].max()+1)
        c_mults = (corners[:,2].min()-1, corners[:,2].max()+1)
        system2 = self.supersize(a_mults, b_mults, c_mults)
        
        # Change system.box.vects to newvects
        system2.box_set(vects=newvects, scale=False)
        
        search_success = False
        for atol in tol:
            
            spos = system2.atoms_prop('pos', scale=True)
            
            # Round atom positions near box boundaries to the boundaries
            spos[np.isclose(spos, 0.0, atol=atol)] = 0.0
            spos[np.isclose(spos, 1.0, atol=atol)] = 1.0
        
            # Identify all atoms whose positions are 0 <= x < 1
            aindex = np.where(((spos[:, 0] >= 0.0) & (spos[:, 0] < 1.0)
                             & (spos[:, 1] >= 0.0) & (spos[:, 1] < 1.0)
                             & (spos[:, 2] >= 0.0) & (spos[:, 2] < 1.0)))
            
            # Check if number of atoms identified matches the expected number
            if len(aindex[0]) == newnatoms:
                search_success = True
                break
        
        if not search_success:
            raise ValueError(f'Filtering failed: {newnatoms} atoms expected, {len(aindex[0])} found')
        
        # Make newsystem by cutting out all atoms in system2 outside boundaries
        newsystem = System(atoms=system2.atoms[aindex], box=system2.box, symbols=self.symbols)
        
        # Return normalized system
        return newsystem.normalize(return_transform=return_transform)
    
    def normalize(self, style='lammps', return_transform=False):
        """
        Normalizes a system's box vectors and atom positions to be compatible
        with simulation codes.
        
        Parameters
        ----------
        style : str, optional
            Indicates the normalization style to use.  Default (and only
            current option) is 'lammps'.
        return_transform : bool, optional
            Indicates if the transformation matrix associated with the
            normalization is returned.  Default value is False.
        Returns
        -------
        newsystem : atomman.System
            A new system that has been normalized.
        transform : np.ndarray
            The transformation matrix associated with the normalization.
            Returned if return_transform is True.
        """
        if style == 'lammps':
            return lmp_normalize(self, return_transform=return_transform)
        else:
            raise ValueError("Unknown style (only 'lammps' is currently supported)")
    
    def dump(self, style, **kwargs):
        """
        Convert a System to another format.
    
        Parameters
        ----------
        style : str
            Indicates the format of the content to dump the atomman.System as.
        kwargs
            Any extra keyword arguments to pass to the underlying dump methods.
            
        Returns
        -------
        str, object or tuple
            Any content returned by the underlying dump methods.
        """
        return dump(style, self, **kwargs)

    def model(self, box_unit=None, prop_name=None, unit=None, prop_unit=None):
        """
        Generates a data model for the System object.

        Parameters
        ----------
        box_unit : str, optional
            Length unit to use for the box. Default value is 'angstrom'.
        prop_name : list, optional
            The Atoms properties to include.  If neither prop_name nor prop_unit
            are given, all system properties will be included.
        unit : list, optional
            Lists the units for each prop_name as stored in the table.  For a
            value of None, no conversion will be performed for that property.  For
            a value of 'scaled', the corresponding table values will be taken in
            box-scaled units.  If neither unit nor prop_units given, pos will be
            given in Angstroms and all other values will not be converted.
        prop_unit : dict, optional
            dictionary where the keys are the property keys to include, and
            the values are units to use. If neither unit nor prop_units given, 
            pos will be given in Angstroms and all other values will not be
            converted.

        Returns
        -------
        DataModelDict.DataModelDict
            A JSON/XML data model for the current System object. 
        """

        # Initialize DataModelDict
        model = DM()
        model['atomic-system'] = DM()

        # Add box
        model['atomic-system']['box'] = self.box.model(length_unit=box_unit)['box']
        
        # Add pbc
        model['atomic-system']['periodic-boundary-condition'] = self.pbc.tolist()
        
        # Add symbols
        for symbol in self.symbols:
            model['atomic-system'].append('atom-type-symbol', symbol)

        # Add masses
        addmasses = False
        for mass in self.masses:
            if mass is not None:
                addmasses = True
                break
        if addmasses:
            for mass in self.masses:
                model['atomic-system'].append('atom-type-mass', mass)
        
        # Add atoms
        model['atomic-system']['atoms'] = amodel = self.atoms.model(prop_name=prop_name,
                                                                    unit=unit, 
                                                                    prop_unit=prop_unit)['atoms']
        
        # Scale properties if needed
        for prop in amodel.aslist('property'):
            if prop['data'].get('unit', None) == 'scaled':
                prop['data'] = uc.model(self.scale(uc.value_unit(prop['data'])), units='scaled') 
        
        return model



from .. import Box, System
from ..tools import vect_angle, miller
from . import dislocation_system_transform

import numpy as np

def dislocation_system_basis(ξ_uvw, slip_hkl, m=None, n=None, box=None, tol=1e-8, maxindex=10,
                             return_hexagonal=False, return_transform=False):
    
    # Set default box to be cubic
    if box is None:
        box = Box()

    # Handle inputs 
    if m is None:
        m = np.array([1,0,0], dtype=float)
    m = np.asarray(m, dtype=float)
    if n is None:
        n = np.array([0,1,0], dtype=float)
    n = np.asarray(n, dtype=float)
    
    # Change ξ_uvw from uvtw if needed
    ξ_uvw = np.asarray(ξ_uvw, dtype=float)
    if ξ_uvw.shape == (4, ):
        ξ_uvw = miller.vector4to3(ξ_uvw)

    # Change slip_hkl from hkil if needed
    slip_hkl = np.asarray(slip_hkl, dtype=float)
    if slip_hkl.shape == (4, ):
        slip_hkl = miller.plane4to3(slip_hkl)

    # Get the transformation matrix for the system
    transform = dislocation_system_transform(ξ_uvw, slip_hkl, m=m, n=n, box=box, tol=1e-8)
    
    # Get orientation settings
    if np.isclose(np.abs(m[0]), 1.0, atol=tol):
        m_cart = transform[0]
    elif np.isclose(np.abs(m[1]), 1.0, atol=tol):
        m_cart = transform[1]
    elif np.isclose(np.abs(m[2]), 1.0, atol=tol):
        m_cart = transform[2]
    else:
        raise ValueError('m must be parallel to a Cartesian axis for this method')
    
    if np.isclose(np.abs(n[0]), 1.0, atol=tol):
        cutboxvector = 'a'
        n_cart = transform[0]
    elif np.isclose(np.abs(n[1]), 1.0, atol=tol):
        cutboxvector = 'b'
        n_cart = transform[1]
    elif np.isclose(np.abs(n[2]), 1.0, atol=tol):
        cutboxvector = 'c'
        n_cart = transform[2]
    else:
        raise ValueError('n must be parallel to a Cartesian axis for this method')
    
    ξ = np.cross(m, n)
    if np.isclose(np.abs(ξ[0]), 1.0, atol=tol):
        ξboxvector = 'a'
    elif np.isclose(np.abs(ξ[1]), 1.0, atol=tol):
        ξboxvector = 'b'
    elif np.isclose(np.abs(ξ[2]), 1.0, atol=tol):
        ξboxvector = 'c'
    else:
        raise ValueError('ξ must be parallel to a Cartesian axis for this method')
    
    # Build gen_vector iterator for testing vectors
    def gen_vector(n):
        for kk in range(0, n+1):
            for sk in [1, -1]:
                k = sk * kk
                for jj in range(0, n+1):
                    for sj in [1, -1]:
                        j = sj * jj
                        for ii in range(0, n+1):
                            for si in [1, -1]:
                                i = si * ii
                                if i==0 and j==0 and k==0:
                                    continue
                                yield np.array([i, j, k], dtype=int)
                                
    # Identify the two crystal vectors that best correspond to m_cart and n_cart
    m_uvw = None
    n_uvw = None    
    m_angle_min = 180
    n_angle_min = 180
    
    for uvw in gen_vector(maxindex):
        cart = np.inner(uvw, box.vects.T)
        m_angle = vect_angle(cart, m_cart)
        n_angle = vect_angle(cart, n_cart)
        
        # Find in-plane vector closest to m
        if np.isclose(np.dot(cart, n_cart), 0.0) and m_angle < m_angle_min:
            m_uvw = uvw
            m_angle_min = m_angle
        
        # Find vector closest to n
        elif n_angle < n_angle_min:
            n_angle_min = n_angle
            n_uvw = uvw
    
    assert m_uvw is not None, 'Failed to find vector near edge component direction'
    assert n_uvw is not None, 'Failed to find vector near slip plane normal'
    
    # Reduce m_uvw and n_uvw if possible
    m_uvw = m_uvw / np.gcd.reduce(np.asarray(m_uvw, dtype=int)) # pylint: disable=no-member
    n_uvw = n_uvw / np.gcd.reduce(np.asarray(n_uvw, dtype=int)) # pylint: disable=no-member
    
    # Orient the uvw sets based on cutboxvector and ξboxvector
    if cutboxvector == 'c':
        if ξboxvector == 'a':
            uvws = np.array([ξ_uvw, m_uvw, n_uvw])
        elif ξboxvector == 'b':
            uvws = np.array([-m_uvw, ξ_uvw, n_uvw])
        else:
            raise RuntimeError('Encountered cutboxvector == ξboxvector: should not be possible!')
    elif cutboxvector == 'b':
        if ξboxvector == 'c':
            uvws = np.array([m_uvw, n_uvw, ξ_uvw])
        elif ξboxvector == 'a':
            uvws = np.array([ξ_uvw, n_uvw, -m_uvw])
        else:
            raise RuntimeError('Encountered cutboxvector == ξboxvector: should not be possible!')
    elif cutboxvector == 'a':
        if ξboxvector == 'b':
            uvws = np.array([n_uvw, ξ_uvw, m_uvw])
        elif ξboxvector == 'c':
            uvws = np.array([n_uvw, -m_uvw, ξ_uvw])
        else:
            raise RuntimeError('Encountered cutboxvector == ξboxvector: should not be possible!')

    # Test if transform from rotation is consistent with transform from above
    system = System(box=box)
    system, test_transform = system.rotate(uvws, return_transform=True)
    
    assert np.allclose(transform, test_transform), 'Crystal system of box does not support the chosen m, n orientation'
    
    if return_hexagonal:
        uvws = miller.vector3to4(uvws)
    
    if return_transform:
    
        return uvws, transform
    else:
        return uvws


# coding: utf-8

# http://www.numpy.org/
import numpy as np

# atomman imports
from . import dvect

def displacement(system_0, system_1, box_reference='final'):
    """
    Compute the displacement vectors between all matching atoms for two systems.
    
    Parameters
    ----------
    system_0 : atomman.System
        The initial system to calculate displacements from.
    system_1 : atomman.System
        The final system to calculate displacements to.
    box_reference : str or None
        Specifies which system's boundary conditions to use.  'initial' uses
        system_0's box and pbc.  'final' uses system_1's box and pbc (Default)
        None computes the straight difference between the positions without
        accounting for periodic boundaries.
    
    Returns
    -------
    numpy.ndarray
        The displacement vectors for all atoms.
        
    Raises
    ------
    ValueError
        If the systems have different numbers of atoms or for invalid
        box_reference values.
    """
    if system_0.natoms != system_1.natoms:
        raise ValueError('systems have different number of atoms')
    
    if box_reference == 'final':
        disp = dvect(system_0.atoms.pos, system_1.atoms.pos, system_1.box, system_1.pbc)
    elif box_reference == 'initial':
        disp = dvect(system_0.atoms.pos, system_1.atoms.pos, system_0.box, system_0.pbc)
    elif box_reference is None:
        disp = system_1.atoms.pos - system_0.atoms.pos
    else:
        raise ValueError("box_reference must be 'final', 'initial', or None")
    
    return disp
    


# coding: utf-8

# http://www.numpy.org/
import numpy as np

# atomman imports
from ... import Atoms, Box, System

#http://pymatgen.org
try:
    import pymatgen as pmg
    has_pmg = True
except:
    has_pmg = False

def load(structure, symbols=None):
    """
    Convert a pymatgen.Structure into an atomman.System.
    
    Parameters
    ----------
    structure : pymatgen.Structure
        A pymatgen representation of a structure.
    symbols : tuple, optional
        Allows the list of element symbols to be assigned during loading.
        Useful if the symbols for the model differ from the standard element
        tags.
    
    Returns
    -------
    system : atomman.System
        A atomman representation of a system.
    """
    
    assert has_pmg, 'pymatgen not imported'
    
    # Get box/lattice information
    box = Box(vects = structure.lattice.matrix)
    
    # Get element information
    all_elements =  np.array([str(symbol) for symbol in structure.species])
    elements, atype = np.unique(all_elements, return_inverse = True)
    if symbols is None:
        symbols = elements

    # Get atomic information
    prop = structure.site_properties
    prop['atype'] = atype + 1
    prop['pos'] = structure.cart_coords
    atoms = Atoms(prop=prop)
    
    # Build system
    system = System(atoms=atoms, box=box, symbols=symbols)
    
    return system


# coding: utf-8
# Standard Python libraries
from collections import OrderedDict
                        
# atomman imports
import atomman.unitconvert as uc
from ... import Atoms, Box, System
from .process_prop_info import process_prop_info, standard_conversions
from ...lammps import style
from .. import load_table
from ...tools import uber_open_rmode

def load(data, symbols=None, lammps_units='metal', prop_name=None,
         table_name=None, shape=None, unit=None, dtype=None,
         prop_info=None, return_prop_info=False):
    """
    Reads in a LAMMPS atomic dump file into a System.
    
    Parameters
    ----------
    data : str or file-like object
        The content, file path or file-like object containing the content to
        read.
    symbols : tuple, optional
        Allows the list of element symbols to be assigned during loading.
    lammps_units : str
        The LAMMPS units option associated with the parameters.  Default value
        is 'metal'.
    prop_name : list, optional
         The Atoms properties to generate.
    table_name : list, optional
        The table column name(s) that correspond to each prop_name.  If
        prop_name, table_name and prop_info are not given, prop_name and
        table_name will be read in from data.
    shape : list, optional
        The shape of each per-atom property.  If not given, will be taken from
        standard LAMMPS parameter names, or left at () for direct 
        property-table conversion.
    unit : list, optional
        Lists the units for each prop_name as stored in the table.  For a
        value of None, no conversion will be performed for that property.  For
        a value of 'scaled', the corresponding table values will be taken in
        box-scaled units.  If not given, all unit values will be set to None
        (i.e. no conversions).
    dtype : list, optional
        Allows for the data type of each property to be explicitly given.
        Values of None will infer the data type from the corresponding
        property values.  If not given, all values will be None.
    prop_info : list of dict, optional
        Structured form of property conversion parameters, in which each
        dictionary in the list corresponds to a single atoms property.  Each
        dictionary must have a 'prop_name' field, and can optionally have
        'table_name', 'shape', 'unit', and 'dtype' fields.
    return_prop_info : bool, optional
        Flag indicating if the full prop_info is to be returned.  Default value
        is False.
        
    Returns
    -------
    system : atomman.System
        The generated system.
    prop_info : list of dict
        The full prop_info detailing the property-table conversion. Returned
        if return_prop_info is True.
    """
    
    lammps_unit = style.unit(lammps_units)
    
    # Initialize parameter values
    pbc = None
    box = None
    natoms = None
    atomsstart = None
    xy = 0.0
    xz = 0.0
    yz = 0.0
    
    readnatoms = False
    readtimestep = False
    bcount = 3
    
    # Read str and files in the same way
    with uber_open_rmode(data) as fp:
        
        # Loop over all lines in fp
        for i, line in enumerate(fp):
            terms = line.decode('UTF-8').split()
            
            # Skip blank lines
            if len(terms) > 0:
                
                # Read number of atoms if time to do so
                if readnatoms:
                    natoms = int(terms[0])
                    readnatoms = False
                
                # Read timestep if time to do so
                elif readtimestep:
                    #timestep = int(terms[0])
                    readtimestep = False
                
                # Read x boundary condition values if time to do so
                elif bcount == 0:
                    xlo = uc.set_in_units(float(terms[0]), lammps_unit['length'])
                    xhi = uc.set_in_units(float(terms[1]), lammps_unit['length'])
                    if len(terms) == 3:
                        xy = uc.set_in_units(float(terms[2]),
                                             lammps_unit['length'])
                    bcount += 1
                
                # Read y boundary condition values if time to do so
                elif bcount == 1:
                    ylo = uc.set_in_units(float(terms[0]), lammps_unit['length'])
                    yhi = uc.set_in_units(float(terms[1]), lammps_unit['length'])
                    if len(terms) == 3:
                        xz = uc.set_in_units(float(terms[2]),
                                             lammps_unit['length'])
                    bcount += 1
                
                # Read z boundary condition values if time to do so
                elif bcount == 2:
                    zlo = uc.set_in_units(float(terms[0]), lammps_unit['length'])
                    zhi = uc.set_in_units(float(terms[1]), lammps_unit['length'])
                    if len(terms) == 3:
                        yz = uc.set_in_units(float(terms[2]),
                                             lammps_unit['length'])
                        
                        # Convert from max, min to hi, lo
                        xlo = xlo - min((0.0, xy, xz, xy + xz))
                        xhi = xhi - max((0.0, xy, xz, xy + xz))
                        ylo = ylo - min((0.0, yz))
                        yhi = yhi - max((0.0, yz))
                    bcount += 1
                
                # Otherwise, only check lines starting with ITEM
                elif terms[0] == 'ITEM:':
                    
                    # ITEM: TIMESTEP indicates it is time to read the timestep
                    if terms[1] == 'TIMESTEP':
                        readtimestep = True
                    
                    # ITEM: NUMBER indicates it is time to read natoms
                    elif terms[1] == 'NUMBER':
                        readnatoms = True
                    
                    # ITEM: BOX gives pbc and indicates it is time to read box parameters
                    elif terms[1] == 'BOX':
                        pbc = [True, True, True]
                        for i in range(3):
                            if terms[i + len(terms) - 3] != 'pp':
                                pbc[i] = False
                        bcount = 0
                        
                    # ITEM: ATOMS gives list of property names and indicates it is time to read atomic values
                    elif terms[1] == 'ATOMS':
                        
                        # Read list of property names
                        name_list = terms[2:]
                        
                        # Create default prop_name and table_name if needed
                        if prop_info is None and prop_name is None:
                            assert table_name is None, 'table_name cannot be given without prop_name'
                            prop_name, table_name = matchprops(name_list)
                            
                        atomsstart = i + 1
    
    # Create system
    box = Box(xlo=xlo, xhi=xhi,
              ylo=ylo, yhi=yhi,
              zlo=zlo, zhi=zhi,
              xy=xy, xz=xz, yz=yz)
    atoms = Atoms(natoms=natoms)
    system = System(box=box, atoms=atoms, pbc=pbc)
    
    # Generate prop_info
    prop_info = process_prop_info(prop_name=prop_name,
                                  table_name=table_name,
                                  shape=shape, unit=unit, dtype=dtype,
                                  prop_info=prop_info,
                                  lammps_units=lammps_units)
    
    # Remove duplicate pos fields
    firstpos = True
    short_prop_info = []
    for pinfo in prop_info:
        if pinfo['prop_name'] in ['pos', 'spos', 'upos', 'supos']:
            if firstpos:
                pinfo['prop_name'] = 'pos'
            else:
                continue
        short_prop_info.append(pinfo)
    
    # Read atoms into system
    system = load_table(data, box=system.box, symbols=symbols, system=system,
                        prop_info=short_prop_info, skiprows=atomsstart,
                        nrows=natoms)
    
    if return_prop_info:
        return system, short_prop_info
    else:
        return system
    
def matchprops(items):
    """
    Takes a list of table_names, pairs them up and matches them to prop_names.
    
    Parameters
    ----------
    items : list
        One dimensional list of all table names.
    
    Returns
    -------
    prop_name : list
        The list of system property names corresponding to items.
    table_name : list
        The list of items paired up and corresponding to the prop_name list.
    """
    
    prop2table = OrderedDict()
    for item in items:
        # Search for prop_name match 
        for sinfo in standard_conversions():
            match = False
            table_names = sinfo['table_name']
            if not isinstance(table_names, list):
                table_names = [table_names]
            if item in table_names:
                match = True
                break
        
        if match is True:
            name = sinfo['prop_name']
        else:
            name = item
        
        if name not in prop2table:
            if match is True:
                for table_name in table_names:
                    assert table_name in items, 'Incomplete propery ' + str(name)
            prop2table[name] = []
        prop2table[name].append(item)
    
    prop_name = list(prop2table.keys())
    table_name = list(prop2table.values())
    for i in range(len(table_name)):
        if len(table_name[i]) == 1:
            table_name[i] = table_name[i][0]
    
    return prop_name, table_name


# coding: utf-8
# Standard Python libraries
from pathlib import Path


import pandas as pd


import numpy as np

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# https://github.com/usnistgov/atomman
import atomman as am

import potentials

from .. import RelaxedCrystal, CrystalPrototype
from ...tools import aslist, screen_input

@property
def relaxed_crystals(self):
    """list or None: Loaded RelaxedCrystal objects"""
    return self.__relaxed_crystals

@property
def relaxed_crystals_df(self):
    """pandas.DataFrame or None: Metadata for loaded RelaxedCrystal objects"""
    return self.__relaxed_crystals_df

def _no_load_relaxed_crystals(self):
    """Initializes properties if load_relaxed_crystals is not called"""
    self.__relaxed_crystals = None
    self.__relaxed_crystals_df = None

def load_relaxed_crystals(self, localpath=None, local=None, remote=None, verbose=False):
    """
    Loads relaxed crystals from the database, first checking localpath, then
    trying to download from host.
    
    Parameters
    ----------
    localpath : str, optional
        Path to a local directory to check for records first.  If not given,
        will check localpath value set during object initialization.  If not
        given or set during initialization, then only the remote database will
        be loaded.
    local : bool, optional
        Indicates if records in localpath are to be loaded.  If not given,
        will use the local value set during initialization.
    remote : bool, optional
        Indicates if the records in the remote database are to be loaded.
        Setting this to be False is useful/faster if a local copy of the
        database exists.  If not given, will use the local value set during
        initialization.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """
    crystals = {}
    
    # Set localpath, local, and remote as given here or during init
    if localpath is None:
        localpath = self.localpath
    if local is None:
        local = self.local
    if remote is None:
        remote = self.remote
    
    # Check localpath first
    if local is True and localpath is not None:
        for fname in Path(localpath, 'relaxed_crystal').glob('*'):
            if fname.suffix in ['.xml', '.json']:
                
                with open(fname, encoding='UTF-8') as f:
                    crystal = RelaxedCrystal(model=f.read())
                crystals[crystal.key] = crystal
                
        if verbose:
            print(f'Loaded {len(crystals)} local relaxed crystals')
    numlocal = len(crystals)

    # Load remote
    if remote is True:
        try:
            records = self.cdcs.query(template='relaxed_crystal')
        except:
            if verbose:
                print('Failed to load relaxed crystals from remote')
        else:
            if verbose:
                print(f'Loaded {len(records)} remote relaxed crystals')
            
            for i in range(len(records)):
                record = records.iloc[i]
                crystal = RelaxedCrystal(record.xml_content)
                if crystal.key not in crystals:
                    crystals[crystal.key] = crystal

            if verbose and len(crystals) > 0 and local:
                print(f' - {len(crystals) - numlocal} new')
    
    # Build relaxed_crystals and relaxed_crystals_df
    if len(crystals) > 0:
        crys = np.array(list(crystals.values()))
        crysdicts = []
        for cry in crystals.values():
            crysdicts.append(cry.asdict())
        
        self.__relaxed_crystals_df = pd.DataFrame(crysdicts).sort_values('key')
        self.__relaxed_crystals = crys[self.relaxed_crystals_df.index]
        self.__relaxed_crystals_df.reset_index(drop=True)
    else:
        if verbose:
            print('No relaxed crystals loaded')
        self.__relaxed_crystals = None
        self.__relaxed_crystals_df = None
        
def get_relaxed_crystals(self, key=None, method='dynamic', standing='good',
                   family=None, parent_key=None, potential=None,
                   potential_LAMMPS_id=None, potential_LAMMPS_key=None, potential_id=None, potential_key=None,
                   symbols=None, natoms=None, natypes=None, keyword=None,
                   verbose=False):
    """
    Get all matching relaxed crystals from the database.
    
    Parameters
    ----------
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    method : str or list or None, optional
        The relaxation method used.  Allowed values are dynamic, static and box.
        Default value is dynamic (the most rigorous relaxation method).  All
        will be loaded if set to None.
    standing : str or list or None, optional
        "good" records are the unique crystals found with the most rigorous
        relaxation, and with known prototypes over DFT structures.  "bad" are
        records filtered out, usually for being duplicates.  Default value is
        "good".  All will be loaded if set to None.
    family : str or atomman.library.CrystalPrototype or list, optional
        The crystal family associated with the relaxed crystal - either crystal
        prototype name or MP/OQMD database entry name.
    parent_key : str or list, optional
        The UUID4 key(s) assigned to the calculation that the record is based
        on.
    potential : atomman.lammps.Potential or list, optional
        A loaded LAMMPS potential object to limit the search by.
    potential_LAMMPS_id : str or list, optional
        The id for a LAMMPS implemented potential to limit the search by.
    potential_LAMMPS_key : str or list, optional
        The UUID4 for a LAMMPS implemented potential to limit the search by.
    potential_id : str or list, optional
        The id for a potential to limit the search by.
    potential_key : str or list, optional
        The UUID4 for a potential to limit the search by.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natypes : int or list, optional
        The number of unique element model symbols in the crystal to limit
        the search by.
    natoms : int or list, optional
        The number of unique atoms in the crystal's unit cell to limit the
        search by.
    keyword : str, optional
        If given, will limit the search to all records that contain the keyword
        substring.  Cannot be combined with any of the above parameters.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    list of RelaxedCrystal
        The matching relaxed crystal records.
    """

    if keyword is not None:
        try:
            assert key is None
            #assert method is None
            #assert standing is None
            assert parent_key is None
            assert family is None
            assert potential is None
            assert potential_LAMMPS_id is None
            assert potential_LAMMPS_key is None
            assert potential_id is None
            assert potential_key is None
            assert symbols is None
            assert natypes is None
            assert natoms is None
        except:
            raise ValueError('keyword cannot be combined with the other search limiting paramters')

    else:
        # Extract potential ids from potential objects
        if potential is not None:
            try:
                assert potential_LAMMPS_id is None
                assert potential_LAMMPS_key is None
                assert potential_id is None
                assert potential_key is None
            except:
                raise ValueError('potential cannot be given with the other potential_* parameters')
            potential_LAMMPS_key = []
            for pot in aslist(potential):
                potential_LAMMPS_key.append(pot.key)

        # Extract family names from prototypes
        if family is not None:
            family = aslist(family)
            for i in range(len(family)):
                if isinstance(family[i], CrystalPrototype):
                    family[i] = family[i].id

    # Check loaded values if available
    if self.relaxed_crystals_df is not None:
        
        if keyword is None:

            def strmatch(series, name, val):
                if val is None:
                    return True
                else:
                    return series[name] in aslist(val)

            def intmatch(series, name, val):
                if val is None:
                    return True
                else:
                    val = aslist(val)
                    for i in range(len(val)):
                        val[i] = int(val[i])
                    return series[name] in val
            
            def listmatch(series, name, val):
                if val is None:
                    return True
                
                elif isinstance(series[name], (list, tuple)):
                    for v in aslist(val):
                        if v not in series[name]:
                            return False
                    return True
                else:
                    return False
            
            crystals = self.relaxed_crystals[self.relaxed_crystals_df.apply(strmatch, args=['key', key], axis=1)
                                             &self.relaxed_crystals_df.apply(strmatch, args=['method', method], axis=1)
                                             &self.relaxed_crystals_df.apply(strmatch, args=['standing', standing], axis=1)
                                             &self.relaxed_crystals_df.apply(strmatch, args=['parent_key', parent_key], axis=1)
                                             &self.relaxed_crystals_df.apply(strmatch, args=['family', family], axis=1)
                                             &self.relaxed_crystals_df.apply(strmatch, args=['potential_LAMMPS_id', potential_LAMMPS_id], axis=1)
                                             &self.relaxed_crystals_df.apply(strmatch, args=['potential_LAMMPS_key', potential_LAMMPS_key], axis=1)
                                             &self.relaxed_crystals_df.apply(strmatch, args=['potential_id', potential_id], axis=1)
                                             &self.relaxed_crystals_df.apply(strmatch, args=['potential_key', potential_key], axis=1)
                                             &self.relaxed_crystals_df.apply(intmatch, args=['natoms', natoms], axis=1)
                                             &self.relaxed_crystals_df.apply(intmatch, args=['natypes', natypes], axis=1)
                                             &self.relaxed_crystals_df.apply(listmatch, args=['symbols', symbols], axis=1)]
        else:
            
            crystals = []
            for crystal in self.relaxed_crystals:
                if keyword.lower() in crystal.asmodel().json().lower():
                    crystals.append(crystal)

        if verbose:
            print(len(crystals), 'matching relaxed crystals found from loaded records')
        return crystals

    # Check remote values if no loaded values
    else:
        if keyword is None:
            # Build Mongoquery
            mquery = {}

            # Add key query
            if key is not None:
                key = aslist(key)
                mquery['relaxed-crystal.key'] = {'$in': key}

            # Add method query
            if method is not None:
                method = aslist(method)
                mquery['relaxed-crystal.method'] = {'$in': method}

            # Add standing query
            if standing is not None:
                standing = aslist(standing)
                mquery['relaxed-crystal.standing'] = {'$in': standing}

            # Add family query
            if family is not None:
                family = aslist(family)
                mquery['relaxed-crystal.system-info.family'] = {'$in': family}

            # Add parent_key query
            if parent_key is not None:
                parent_key = aslist(parent_key)
                mquery['relaxed-crystal.system-info.parent_key'] = {'$in': parent_key}
            
            # Add potential_LAMMPS_id query
            if potential_LAMMPS_id is not None:
                potential_LAMMPS_id = aslist(potential_LAMMPS_id)
                mquery['relaxed-crystal.potential-LAMMPS.id'] = {'$in': potential_LAMMPS_id}

            # Add potential_LAMMPS_key query
            if potential_LAMMPS_key is not None:
                potential_LAMMPS_key = aslist(potential_LAMMPS_key)
                mquery['relaxed-crystal.potential-LAMMPS.key'] = {'$in': potential_LAMMPS_key}

            # Add potential_id query
            if potential_id is not None:
                potential_id = aslist(potential_id)
                mquery['relaxed-crystal.potential-LAMMPS.potential.id'] = {'$in': potential_id}

            # Add potential_key query
            if potential_key is not None:
                potential_key = aslist(potential_key)
                mquery['relaxed-crystal.potential-LAMMPS.potential.key'] = {'$in': potential_key}

            # Add natoms query
            if natoms is not None:
                natoms = aslist(natoms)
                for i in range(len(natoms)):
                    natoms[i] = int(natoms[i])
                mquery['relaxed-crystal.atomic-system.atoms.natoms'] = {'$in': natoms}

            # Add natypes query
            if natypes is not None:
                natypes = aslist(natypes)
                mquery['$or'] = []
                for natype in natypes:
                    if natype == 1:
                        mquery['$or'].append({"relaxed-crystal.atomic-system.atom-type-symbol":{"$not":{"$type":"array"}}})
                    else:
                        mquery['$or'].append({"relaxed-crystal.atomic-system.atom-type-symbol":{"$size":natype}})

            # Add symbol query
            if symbols is not None:
                symbols = aslist(symbols)
                mquery['relaxed-crystal.atomic-system.atom-type-symbol'] = {'$all': symbols}
            
            matches = self.cdcs.query(template='relaxed_crystal', mongoquery=mquery)

        else:
            matches = self.cdcs.query(template='relaxed_crystal', keyword=keyword)

        if verbose:
            print(len(matches), 'matching relaxed crystals found from remote database')

        if len(matches) > 0:
            def makecrystals(series):
                return RelaxedCrystal(model=series.xml_content)
            crystals = matches.apply(makecrystals, axis=1).values

            crystals_df = []
            for crystal in crystals:
                crystals_df.append(crystal.asdict())
            crystals_df = pd.DataFrame(crystals_df)

            return crystals[crystals_df.sort_values('cohesive_energy').index]
            
        else:
            return np.array([])

def get_relaxed_crystal(self, key=None, method='dynamic', standing='good',
                   family=None, parent_key=None, potential=None,
                   potential_LAMMPS_id=None, potential_LAMMPS_key=None, potential_id=None, potential_key=None,
                   symbols=None, natoms=None, natypes=None, keyword=None,
                   verbose=False):
    """
    Gets a single matching relaxed crystal from the database. If multiple
    matches are found, a selection menu will appear.
    
    Parameters
    ----------
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    method : str or list or None, optional
        The relaxation method used.  Allowed values are dynamic, static and box.
        Default value is dynamic (the most rigorous relaxation method).  All
        will be loaded if set to None.
    standing : str or list or None, optional
        "good" records are the unique crystals found with the most rigorous
        relaxation, and with known prototypes over DFT structures.  "bad" are
        records filtered out, usually for being duplicates.  Default value is
        "good".  All will be loaded if set to None.
    family : str or atomman.library.CrystalPrototype or list, optional
        The crystal family associated with the relaxed crystal - either crystal
        prototype name or MP/OQMD database entry name.
    parent_key : str or list, optional
        The UUID4 key(s) assigned to the calculation that the record is based
        on.
    potential : atomman.lammps.Potential or list, optional
        A loaded LAMMPS potential object to limit the search by.
    potential_LAMMPS_id : str or list, optional
        The id for a LAMMPS implemented potential to limit the search by.
    potential_LAMMPS_key : str or list, optional
        The UUID4 for a LAMMPS implemented potential to limit the search by.
    potential_id : str or list, optional
        The id for a potential to limit the search by.
    potential_key : str or list, optional
        The UUID4 for a potential to limit the search by.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natypes : int or list, optional
        The number of unique element model symbols in the crystal to limit
        the search by.
    natoms : int or list, optional
        The number of unique atoms in the crystal's unit cell to limit the
        search by.
    keyword : str, optional
        If given, will limit the search to all records that contain the keyword
        substring.  Cannot be combined with any of the above parameters.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    RelaxedCrystal
        A selected relaxed crystal record.

    Raises
    ------
    ValueError
        If no matching records are identified.
    """
    crystals = self.get_relaxed_crystals(key=key, method=method, standing=standing,
                   family=family, parent_key=parent_key, potential=potential,
                   potential_LAMMPS_id=potential_LAMMPS_id, potential_LAMMPS_key=potential_LAMMPS_key,
                   potential_id=potential_id, potential_key=potential_key,
                   symbols=symbols, natoms=natoms, natypes=natypes, keyword=keyword,
                   verbose=verbose)

    # Check number of matches and select
    if len(crystals) == 1:
        return crystals[0]
    
    elif len(crystals) > 1:
        print('Multiple matching relaxed crystals found.')

        # Create list header
        header = '#  family               symbols  alat    Ecoh    '
        if method is None:
            header += 'method  '
        if standing is None:
            header += 'standing'
        print(header)
        
        # Generate options list
        for i in range(len(crystals)):
            crystal = crystals[i]
            
            row = f'{i+1:2} {crystal.family:20.20} {"".join(crystal.ucell.symbols):8.8} {crystal.ucell.box.a:7.4f} {crystal.cohesive_energy:7.4f} '
            
            if method is None:
                row += f'{crystal.method:7.7} '
            if standing is None:
                row += f'{crystal.standing:4.4}'
            print(row)

        choice = int(screen_input('Select which one:'))
        return crystals[choice-1]
    
    else:
        raise ValueError('No matching relaxed crystals found')

def download_relaxed_crystals(self, localpath=None, crystals=None, standing='good',
                              format='json', indent=None, verbose=False):
    """
    Download citation records from the remote and save to localpath.
    
    Parameters
    ----------
    localpath : path-like object, optional
        Path to a local directory where the files will be saved to.  If not
        given, will use the localpath value set during object initialization.
    crystals : list of RelaxedCrystals, optional
        A list of relaxed crystals to download. If not given, all (good)
        crystals will be downloaded.
    standing : str or None, optional
        By default, only records with standing == good will be downloaded.
        Setting this to None will download all crystals.
    format : str, optional
        The file format to save the record files as.  Allowed values are
        'xml' and 'json' (default).
    indent : int, optional
        The indentation spacing size to use for the locally saved record files.
        If not given, the JSON/XML content will be compact.  Ignored if format
        is 'bib'.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """

    template = 'relaxed_crystal'

    # Handle localpath value
    if localpath is None:
        localpath = self.localpath
    if localpath is None:
        raise ValueError('No local path set to save files to')
    
    # Check format value
    format = format.lower()
    allowed_formats = ['xml', 'json']
    if format not in allowed_formats:
        raise ValueError("Format must be 'xml' or 'json'")

    # Create save directory if needed
    save_directory = Path(localpath, template)
    if not save_directory.is_dir():
        save_directory.mkdir(parents=True)

    for fmt in allowed_formats:
        if fmt != format:
            numexisting = len([fname for fname in save_directory.glob(f'*.{fmt}')])
            if numexisting > 0:
                raise ValueError(f'{numexisting} records of format {fmt} already saved locally')

    # Download if needed
    if crystals is None:
        if standing is not None:
            mquery = {}
            mquery['relaxed-crystal.standing'] = {'$in': standing}
        else:
            mquery = None

        self.download_records(template=template, localpath=localpath, mongoquery=mquery,
                              format=format, indent=indent, verbose=verbose)
        
    else:
        crystals = aslist(crystals)

        for crystal in crystals:
            fname = Path(save_directory, f'{crystal.id}.{format}')
            if format == 'xml':
                with open(fname, 'w', encoding='UTF-8') as f:
                    crystal.asmodel().xml(fp=f, indent=indent)
            elif format == 'json':
                with open(fname, 'w', encoding='UTF-8') as f:
                    crystal.asmodel().json(fp=f, indent=indent)
        if verbose:
            print(f'Copied {len(crystals)} records to localpath')

def upload_relaxed_crystal(self, crystal, workspace=None, verbose=False):
    raise NotImplementedError('To be done...')


# coding: utf-8
# Standard Python libraries
from copy import deepcopy

# atomman imports
from ...lammps import style
from ...tools import indexstr

def process_prop_info(prop_name=None, table_name=None, shape=None, unit=None,
                      dtype=None, prop_info=None, lammps_units='metal'):
    """Handles common setting of prop_info terms"""
    
    if prop_info is not None:
        # Check that competing parameters are not given
        try:
            assert prop_name is None
            assert table_name is None
            assert shape is None
            assert unit is None
            assert dtype is None
        except:
            raise ValueError('prop_info cannot be given with prop_name, table_name, shape, unit, or dtype')
        prop_info = deepcopy(prop_info)
    
    # Combine list parameters into prop_info
    else:
        
        # Check that prop_name is given
        if prop_name is None:
            raise ValueError('prop_info or prop_name is required.')
        numprops = len(prop_name)
        
        # Check that sizes match
        try:
            assert table_name is None or len(table_name) == numprops
            assert shape is None or len(shape) == numprops
            assert unit is None or len(unit) == numprops
            assert dtype is None or len(dtype) == numprops
        except:
            raise ValueError('any of prop_name, table_name, shape, unit, and dtype given must be the same length')
        
        # Build prop_info
        prop_info = []
        for i in range(len(prop_name)):
            
            # Search for prop_name match 
            for sinfo in standard_conversions(lammps_units):
                match = False
                if sinfo['prop_name'] == prop_name[i]:
                    match = True
                    break
            
            pinfo = {}
            pinfo['prop_name'] = prop_name[i]
            
            if table_name is not None:
                pinfo['table_name'] = table_name[i]
            elif match is True and 'table_name' in sinfo:
                pinfo['table_name'] = sinfo['table_name']
            
            if shape is not None:
                pinfo['shape'] = shape[i]
            elif match is True and 'shape' in sinfo:
                pinfo['shape'] = sinfo['shape']
                
            if unit is not None:
                pinfo['unit'] = unit[i]
            elif match is True and 'unit' in sinfo:
                pinfo['unit'] = sinfo['unit']
                
            if dtype is not None:
                pinfo['dtype'] = dtype[i]
            elif match is True and 'dtype' in sinfo:
                pinfo['dtype'] = sinfo['dtype']
            
            prop_info.append(pinfo)
    
    # Fill in missing default values
    for prop in prop_info:
        
        # prop_name is required
        if 'prop_name' not in prop:
            raise KeyError('prop_name required for each entry in prop_info')
        
        # Generate table_name from prop_name and shape
        if 'table_name' not in prop:
            prop['shape'] = prop.get('shape', ())
            prop['table_name'] = []
            for index, istr in indexstr(prop['shape']): #pylint: disable=unused-variable
                prop['table_name'].append(prop['prop_name'] + istr)
        # Make certain table_name is a list
        else:
            try:
                assert not isinstance(prop['table_name'], str)
                prop['table_name'] = list(prop['table_name'])
            except:
                prop['table_name'] = [prop['table_name']]
        
        # Take shape as length of table_name
        if 'shape' not in prop:
            numtnames = len(prop['table_name'])
            if numtnames == 1:
                prop['shape'] = ()
            else:
                prop['shape'] = (numtnames, )
        # Make certain shape is a tuple
        else:
            try:
                prop['shape'] = tuple(prop['shape'])
            except:
                prop['shape'] = (prop['shape'],)
        
        # Default unit and dtype values are None
        prop['unit'] = prop.get('unit', None)
        prop['dtype'] = prop.get('dtype', None)
    
    return prop_info

def standard_conversions(lammps_units='metal'):
    """
    Gives pre-defined conversion information for standard LAMMPS variables.
    """
    
    lammps_unit = style.unit(lammps_units)
    
    return [{"prop_name": "atom_id",
             "table_name": "id"},
            
            {"prop_name": "m_id",
             "table_name": "mol"},
            
            {"prop_name": "p_id",
             "table_name": "proc"},
            
            {"prop_name": "p_id_plus1",
             "table_name": "procp1"},
            
            {"prop_name": "atype",
             "table_name": "type"},
            
            {"prop_name": "element",
             "table_name": "element"},
            
            {"prop_name": "mass",
             "table_name": "mass",
             "unit": lammps_unit['mass']},
            
            {"prop_name": "pos",
             "table_name": ["x", "y", "z"],
             "unit": lammps_unit['length']},
            
            {"prop_name": "spos",
             "table_name": ["xs", "ys", "zs"],
             "unit": 'scaled'},
            
            {"prop_name": "upos",
             "table_name": ["xu", "yu", "zu"],
             "unit": lammps_unit['length']},
            
            {"prop_name": "supos",
             "table_name": ["xsu", "ysu", "zsu"],
             "unit": 'scaled'},
            
            {"prop_name": "boximage",
             "table_name": ["ix", "iy", "iz"],
             "unit": 'scaled'},
            
            {"prop_name": "velocity",
             "table_name": ["vx", "vy", "vz"],
             "unit": lammps_unit['velocity']},
            
            {"prop_name": "force",
             "table_name": ["fx", "fy", "fz"],
             "unit": lammps_unit['force']},
            
            {"prop_name": "charge",
             "table_name": "q",
             "unit": lammps_unit['charge']},
            
            {"prop_name": "mu",
             "table_name": ["mux", "muy", "muz"],
             "unit": lammps_unit['dipole']},
            
            {"prop_name": "mu_mag",
             "table_name": "mu",
             "unit": lammps_unit['dipole']},
            
            {"prop_name": "radius",
             "table_name": "radius",
             "unit": lammps_unit['length']},
            
            {"prop_name": "diameter",
             "table_name": "diameter",
             "unit": lammps_unit['length']},
            
            {"prop_name": "ang_velocity",
             "table_name": ["omegax", "omegay", "omegaz"],
             "unit": lammps_unit['ang-vel']},
            
            {"prop_name": "ang_momentum",
             "table_name": ["angmomx", "angmomy", "angmomz"],
             "unit": lammps_unit['ang-mom']},
            
            {"prop_name": "torque",
             "table_name": ["tqx", "tqy", "tqz"],
             "unit": lammps_unit['force'] + '*' + lammps_unit['length']}]


# coding: utf-8

def rungekutta(ratefxn, coord, timestep, **kwargs):
    """
    Performs Runge-Kutta ODE integration for a timestep.
    
    Parameters
    ----------
    ratefxn : function
        The rate function to use.
    coord : array-like object
        The coordinate(s) of the last timestep.
    timestep : float
        The timestep value to use.
    **kwargs : any
        Any extra keyword parameters to pass on to ratefxn.
    
    Returns
    -------
    array-like object
        The coordinate(s) moved forward by timestep.
    """
    
    k1 = timestep * ratefxn(coord, **kwargs)
    k2 = timestep * ratefxn(coord - 0.5 * k1, **kwargs)
    k3 = timestep * ratefxn(coord - 0.5 * k2, **kwargs)
    k4 = timestep * ratefxn(coord - k3, **kwargs)
    
    return coord + k1 / 6 + k2 / 3 + k3 / 3 + k4 / 6


from copy import deepcopy
from DataModelDict import DataModelDict as DM

import atomman.unitconvert as uc
from .. import System
from ..tools import identifyfamily
class RelaxedCrystal():
    
    def __init__(self, model):
        self.load(model=model)

    def __str__(self):
        return f'Relaxed Crystal {self.key}'

    @property
    def key(self):
        return self.__key

    @property
    def method(self):
        return self.__method
    
    @property
    def standing(self):
        return self.__standing
    
    @property
    def family(self):
        return self.__family

    @property
    def parent_key(self):
        return self.__parent_key

    @property
    def potential_LAMMPS_id(self):
        return self.__potential_LAMMPS_id

    @property
    def potential_LAMMPS_key(self):
        return self.__potential_LAMMPS_key

    @property
    def potential_id(self):
        return self.__potential_id

    @property
    def potential_key(self):
        return self.__potential_key

    @property
    def cohesive_energy(self):
        return self.__cohesive_energy

    @property
    def ucell(self):
        return self.__ucell

    def load(self, model):
        
        self.__model = model = DM(model)
        
        crystal = model['relaxed-crystal']
        self.__key = crystal['key']
        self.__method = crystal['method']
        self.__standing = crystal['standing']

        self.__family = crystal['system-info']['family']
        self.__parent_key = crystal['system-info']['parent_key']
        
        self.__potential_LAMMPS_id = crystal['potential-LAMMPS']['id']
        self.__potential_LAMMPS_key = crystal['potential-LAMMPS']['key']
        self.__potential_id = crystal['potential-LAMMPS']['potential']['id']
        self.__potential_key = crystal['potential-LAMMPS']['potential']['key']
        self.__cohesive_energy = uc.value_unit(crystal['cohesive-energy'])
        
        self.__ucell = System(model=model)

    def asmodel(self):
        
        return deepcopy(self.__model)

    def asdict(self):
        """
        Converts the structured content to a simpler dictionary.
        
        Returns
        -------
        dict
            A dictionary representation of the record's content.
        """
        params = {}
        params['key'] = self.key
        params['method'] = self.method
        params['standing'] = self.standing
        params['family'] = self.family
        params['parent_key'] = self.parent_key
        
        params['potential_LAMMPS_id'] = self.potential_LAMMPS_id
        params['potential_LAMMPS_key'] = self.potential_LAMMPS_key
        params['potential_id'] = self.potential_id
        params['potential_key'] = self.potential_key
        
        params['cohesive_energy'] = self.cohesive_energy

        params['ucell'] = ucell = self.ucell
        params['crystal_family'] = identifyfamily(ucell.box)
        params['natypes'] = ucell.natypes
        params['symbols'] = ucell.symbols
        
        params['a'] = ucell.box.a
        params['b'] = ucell.box.b
        params['c'] = ucell.box.c
        params['alpha'] = ucell.box.alpha
        params['beta'] = ucell.box.beta
        params['gamma'] = ucell.box.gamma
        params['natoms'] = ucell.natoms


        
        return params


# coding: utf-8
# Standard Python libraries
from copy import deepcopy

# http://www.numpy.org/
import numpy as np

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

from . import (dislocation_system_basis, solve_volterra_dislocation,
               dislocation_array)
from ..tools import miller, vect_angle
from ..region import PlaneSet, Cylinder

def boolean(value):
    """
    Allows conversion of strings to Booleans.
    
    Parameters
    ----------
    value : str or bool
        If str, then 'true' and 't' become True and 'false' and 'f' become
        false. If bool, simply return the value.
        
    Returns
    -------
    bool
        Equivalent bool of value.
        
    Raises
    ------
    ValueError
        If value is unrecognized.
    """
    
    # Pass Boolean values through without changing
    if value is True:
        return True
    elif value is False:
        return False
    
    # Convert strings
    elif value.lower() in ['true', 't']:
        return True
    elif value.lower() in ['false', 'f']:
        return False
    
    # Issue error for invalid string
    else:
        raise ValueError('Invalid Boolean string')

class Dislocation():
    
    def __init__(self, ucell, C, burgers, ξ_uvw, slip_hkl, m=[0,1,0],
                 n=[0,0,1], shift=None, shiftindex=None, shiftscale=None, tol=1e-8):
        """
        Class initializer.  Solves the dislocation solution and rotates the
        given unit cell to the proper orientation.
        
        Parameters
        ----------
        ucell : atomman.System
            The unit cell to use as the seed for generating the dislocation
            monopole system.
        C : atomman.ElasticConstants
            The elastic constants associated with the bulk crystal structure
            for ucell.
        burgers : array-like object
            The dislocation's Burgers vector given as a Miller or
            Miller-Bravais vector relative to ucell.
        ξ_uvw : array-like object
            The dislocation's line direction given as a Miller or
            Miller-Bravais vector relative to ucell.
        slip_hkl : array-like object
            The dislocation's slip plane given as a Miller or Miller-Bravais
            plane relative to ucell.
        m : array-like object, optional
            The m unit vector for the dislocation solution.  m, n, and ξ
            (dislocation line) should be right-hand orthogonal.  Default value
            is [0,1,0] (y-axis).
        n : array-like object, optional
            The n unit vector for the dislocation solution.  m, n, and ξ
            (dislocation line) should be right-hand orthogonal.  Default value
            is [0,0,1] (z-axis). n is normal to the dislocation slip plane.
        shift : float, optional
            A rigid body shift to apply to the rotated cell prior to inserting
            the dislocation.  Should be selected such that the ideal slip plane
            does not correspond to any atomic planes.  Is taken as absolute if
            shiftscale is False, or relative to the rotated cell's box vectors
            if shiftscale is True.  Cannot be given with shiftindex.  If
            neither shift nor shiftindex is given then shiftindex = 0 is used.
        shiftindex : float, optional
            The index of the identified optimum shifts based on the rotated
            cell to use.  Different values allow for the selection of different
            atomic planes neighboring the slip plane.  Note that shiftindex
            values only apply shifts normal to the slip plane; best shifts for
            non-planar dislocations (like bcc screw) may also need a shift in
            the slip plane.  Cannot be given with shiftindex.  If neither shift
            nor shiftindex is given then shiftindex = 0 is used.
        shiftscale : bool, optional
            If False (default), a given shift value will be taken as absolute
            Cartesian.  If True, a given shift will be taken relative to the
            rotated cell's box vectors.
        tol : float
            A cutoff tolerance used with obtaining the dislocation solution.
            Only needs to be changed if there are issues with obtaining a
            solution.
        """
        # Generate the dislocation solution
        self.__dislsol = solve_volterra_dislocation(C, burgers, ξ_uvw=ξ_uvw,
                                                    slip_hkl=slip_hkl, m=m, n=n,
                                                    box=ucell.box, tol=tol)
        self.__transform = self.dislsol.transform
        
        # Compute the uvws and transformation matrix corresponding to the
        # dislocation solution
        self.__uvws = dislocation_system_basis(ξ_uvw, slip_hkl, m=m, n=n,
                                        box=ucell.box, tol=tol)
        
        # Rotate ucell to correspond to the dislocation solution
        self.__rcell = ucell.rotate(self.uvws)
        self.__ucell = ucell
        
        # Identify the lattice vector aligned with ξ
        a_angle = vect_angle(self.rcell.box.avect, self.dislsol.ξ)
        b_angle = vect_angle(self.rcell.box.bvect, self.dislsol.ξ)
        c_angle = vect_angle(self.rcell.box.cvect, self.dislsol.ξ)
        
        if np.isclose(a_angle, 0.0, atol=tol, rtol=0.0) or np.isclose(a_angle, 180.0, atol=tol, rtol=0.0):
            self.__lineindex = 0
        elif np.isclose(b_angle, 0.0, atol=tol, rtol=0.0) or np.isclose(b_angle, 180.0, atol=tol, rtol=0.0):
            self.__lineindex = 1
        elif np.isclose(c_angle, 0.0, atol=tol, rtol=0.0) or np.isclose(c_angle, 180.0, atol=tol, rtol=0.0):
            self.__lineindex = 2
        else:
            raise ValueError('A box vector of the rotated system must correspond with m X n = ξ')
        
        # Identify the lattice vector that is not in the slip plane
        adot = np.dot(self.rcell.box.avect, self.dislsol.n)
        bdot = np.dot(self.rcell.box.bvect, self.dislsol.n)
        cdot = np.dot(self.rcell.box.cvect, self.dislsol.n)
        
        if np.isclose(adot, 0.0, atol=tol, rtol=0.0) and np.isclose(bdot, 0.0, atol=tol, rtol=0.0):
            self.__cutindex = 2
        elif np.isclose(adot, 0.0, atol=tol, rtol=0.0) and np.isclose(cdot, 0.0, atol=tol, rtol=0.0):
            self.__cutindex = 1
        elif np.isclose(bdot, 0.0, atol=tol, rtol=0.0) and np.isclose(cdot, 0.0, atol=tol, rtol=0.0):
            self.__cutindex = 0
        else:
            raise ValueError('Only one box vector of the rotated system must have a component along n')
        
        # Define out of plane unit vector 
        ovect = np.zeros(3)
        ovect[self.cutindex] = 1.0
        
        # Get out of plane width
        rcellwidth = self.rcell.box.vects[self.cutindex, self.cutindex]
        
        # Get the unique coordinates normal to the plane
        pos = self.rcell.atoms.pos
        numdec = - int(np.floor(np.log10(tol)))
        coords = np.unique(pos[:, self.cutindex].round(numdec))

        # Add periodic replica if missing
        if not np.isclose(coords[-1] - coords[0], rcellwidth, rtol=0.0, atol=tol):
            coords = np.append(coords, coords[0] + rcellwidth)
            
        # Compute the shifts
        relshifts = rcellwidth - (coords[1:] + coords[:-1]) / 2
        relshifts[relshifts > rcellwidth] -= rcellwidth
        relshifts[relshifts < 0.0] += rcellwidth
        self.__shifts = np.outer(np.sort(relshifts), ovect)
        
        # Handle shift parameters
        if shift is not None:
            if shiftindex is not None:
                raise ValueError('shift and shiftindex cannot both be given')
            if shiftscale is True:
                self.__shift = miller.vector_crystal_to_cartesian(shift, self.rcell.box)
            else:
                self.__shift = np.asarray(shift)
                assert self.__shift.shape == (3,) 
        elif shiftindex is not None:
            self.__shift = self.shifts[shiftindex]
        else:
            self.__shift = self.shifts[0]

        # Set base_system and disl_system to None
        self.__base_system = None
        self.__disl_system = None

    @classmethod
    def fromref(cls, ucell, C, model, tol=1e-8):
        """
        Initializes a Dislocation object based on pre-defined dislocation
        parameters from a reference record.

        Parameters
        ----------
        ucell : atomman.System
            The unit cell to use as the seed for generating the dislocation
            monopole system.
        C : atomman.ElasticConstants
            The elastic constants associated with the bulk crystal structure
            for ucell.
        model : str, file-like object or DataModelDict
            The reference record containing dislocation parameters to use.
        tol : float
            A cutoff tolerance used with obtaining the dislocation solution.
            Only needs to be changed if there are issues with obtaining a
            solution.
        """
        # Load the record
        model = DM(model).find('dislocation')

        # Extract the dislocation parameters
        cp = model['calculation-parameter']
        slip_hkl = np.array(cp['slip_hkl'].split(), dtype=int)
        ξ_uvw = np.array(cp['ξ_uvw'].split(), dtype=int)
        burgers = np.array(cp['burgers'].split(), dtype=float)
        m = np.array(cp.get('m', '0 1 0').split(), dtype=float)
        n = np.array(cp.get('n', '0 0 1').split(), dtype=float)

        shift = cp.get('shift', None)
        shiftindex = cp.get('shiftindex', None)
        shiftscale = boolean(cp.get('shiftscale', False))
        
        if shift is not None:
            shift = np.array(shift.split(), dtype='float')
        elif shiftindex is not None:
            shiftindex = int(shiftindex)
            
        return cls(ucell, C, burgers, ξ_uvw, slip_hkl, m=m, n=n, shift=shift,
                   shiftindex=shiftindex, shiftscale=shiftscale, tol=tol)
    
    @property
    def dislsol(self):
        """atomman.defect.VolterraDislocation : The elastic dislocation solution"""
        return self.__dislsol
    
    @property
    def uvws(self):
        """numpy.NDArray : The 3x3 array of uvw Miller vectors used to rotate ucell to rcell"""
        return self.__uvws
    
    @property
    def transform(self):
        """numpy.NDArray : The 3x3 Cartesian transformation matrix associated with rotating from ucell to rcell"""
        return self.__transform
    
    @property
    def ucell(self):
        """atomman.System : The crystal unit cell used as the basis for constructing the dislocation system"""
        return self.__ucell
    
    @property
    def rcell(self):
        """atomman.System : The cell associated with rotating ucell to coincide with the dislocation solution"""
        return self.__rcell
    
    @property
    def lineindex(self):
        """int : The index of the box vector that coincides with the dislocation line: 0=a, 1=b, 2=c"""
        return self.__lineindex
    
    @property
    def cutindex(self):
        """int : The index of the box vector that is not within the slip plane: 0=a, 1=b, 2=c"""
        return self.__cutindex
    
    @property
    def shifts(self):
        """list : All identified shifts that will place the slip plane halfway between atomic planes"""
        return self.__shifts
    
    @property
    def shift(self):
        """numpy.NDArray : The particular shift value that will be or was used to construct the dislocation system"""
        return self.__shift

    @property
    def base_system(self):
        """atomman.System : The "perfect crystal" reference system associated with the dislocation system"""
        if self.__base_system is not None:
            return self.__base_system
        else:
            raise ValueError('base_system not built yet: must call monopole() or periodicarray() first')

    @property
    def disl_system(self):
        """atomman.System : The generated dislocation system"""
        if self.__disl_system is not None:
            return self.__disl_system
        else:
            raise ValueError('disl_system not built yet: must call monopole() or periodicarray() first')

    def box_boundary(self, box, width):
        """
        Constructs a shape associated with the box-style boundary region.  Used
        by the monopole() generation method.  The returned shape will encompass
        all atoms except those within width distance of the two non-periodic 
        surfaces.

        Parameters
        ----------
        box : atomman.Box
            The box associated with the full (base) system.
        width : float
            The width of the boundary region

        Returns
        -------
        atomman.region.PlaneSet
            The Shape object excluding the boundary region
        """
        # Get plane shapes for the two non-periodic directions
        planes = []
        for i in range(3):
            if i == self.lineindex:
                continue
            planes.append(box.planes[i])
            planes.append(box.planes[i+3])

        # Shift plane points by width in the plane normal directions
        for plane in planes:
            plane.point -= width * plane.normal

        # Create and return shape
        return PlaneSet(planes)
    
    def array_boundary(self, box, width):
        """
        Constructs a shape associated with the boundary regions used by the
        periodicarray() generation method.  The returned shape will encompass
        all atoms except those within width distance of the non-periodic 
        surface.

        Parameters
        ----------
        box : atomman.Box
            The box associated with the full (base) system.
        width : float
            The width of the boundary region

        Returns
        -------
        atomman.region.PlaneSet
            The Shape object excluding the boundary region
        """
        # Get plane shapes for the two non-periodic directions
        planes = [box.planes[self.cutindex], box.planes[self.cutindex + 3]]
        
        # Shift plane points by width in the plane normal directions
        for plane in planes:
            plane.point -= width * plane.normal

        # Create and return shape
        return PlaneSet(planes)

    def cylinder_boundary(self, box, width):
        """
        Constructs a shape associated with the cylinder-style boundary region.
        Used by the monopole() generation method.  The returned shape will
        encompass a cylinder of atoms centered around the dislocation line
        leaving a boundary region that will be at least width wide everywhere.

        Parameters
        ----------
        box : atomman.Box
            The box associated with the full (base) system.
        width : float
            The minimum width of the boundary region.

        Returns
        -------
        atomman.region.Cylinder
            The Shape object excluding the boundary region
        """
        # Reduce the problem to 2D: solution independent of ξ position
        mn = np.array([self.dislsol.m, self.dislsol.n])
        vect1 = mn.dot(box.vects[self.lineindex - 2])
        vect2 = mn.dot(box.vects[self.lineindex - 1])
        origin = mn.dot(box.origin)

        # Compute normal vectors to box vectors
        normal_vect1 = np.array([-vect1[1], vect1[0]])
        normal_vect2 = np.array([vect2[1], -vect2[0]])
        normal_vect1 = normal_vect1 / np.linalg.norm(normal_vect1)
        normal_vect2 = normal_vect2 / np.linalg.norm(normal_vect2)

        def line(p1, p2):
            """
            Defines a 2D line as used by the intersection function
            """
            A = (p1[1] - p2[1])
            B = (p2[0] - p1[0])
            C = (p1[0]*p2[1] - p2[0]*p1[1])
            return A, B, -C

        def intersection(L1, L2):
            """
            Identifies the (x,y) coordinates where two 2D lines intersect
            """
            D  = L1[0] * L2[1] - L1[1] * L2[0]
            Dx = L1[2] * L2[1] - L1[1] * L2[2]
            Dy = L1[0] * L2[2] - L1[2] * L2[0]
            if D != 0:
                x = Dx / D
                y = Dy / D
                return x, y
            else:
                return False

        # Define normal lines as originating at (0,0)
        normal_line_1 = line([0,0], normal_vect1)
        normal_line_2 = line([0,0], normal_vect2)

        # Define boundary lines based on 2D box corners
        bound_bot1 = line(origin, origin + vect1)
        bound_bot2 = line(origin, origin + vect2)
        bound_top1 = line(origin + vect2, origin + vect2 + vect1)
        bound_top2 = line(origin + vect1, origin + vect1 + vect2)

        # Identify intersection points between normal lines and the boundary lines
        intersections = np.array([intersection(normal_line_1, bound_bot1),
                                  intersection(normal_line_2, bound_bot2),
                                  intersection(normal_line_1, bound_top1),
                                  intersection(normal_line_2, bound_top2)])

        # Find distance between (0,0) and the closest intercept
        smallest = np.min(np.linalg.norm(intersections, axis=1))

        # Radius = smallest distance minus the boundary width
        radius =  smallest - width

        # Axis is along the line direction and includes point (0,0,0)  
        center1 = np.zeros(3)
        center2 = box.vects[self.lineindex]

        return Cylinder(center1, center2, radius, endcaps=False)
    
    def monopole(self, sizemults=None, amin=0.0, bmin=0.0, cmin=0.0,
                 shift=None, shiftindex=None, shiftscale=False,
                 boundaryshape='cylinder', boundarywidth=0.0,
                 boundaryscale=False, return_base_system=False):
        """
        Constructs a dislocation monopole atomic configuration containing a
        single perfectly straight dislocation. The resulting system will be
        periodic along the box vector direction that corresponds to the
        dislocation's line direction, and non-periodic in the other two box
        vector directions.  Boundary atoms near the two free surfaces will be
        identified by changing their atype values making it easy to identify
        them later for assigning different boundary conditions.
        
        Parameters
        ----------
        sizemults : tuple, optional
            The size multipliers to use when generating the system.  Values are
            limited to being positive integers.  The multipliers for the two
            non-periodic directions must be even.  If not given, the default
            multipliers will be 2 for the non-periodic directions and 1 for the
            periodic direction.
        amin : float, optional
            A minimum thickness to use for the a box vector direction of the
            final system.  Default value is 0.0.  For the non-periodic
            directions, the resulting vector multiplier will be even.  If both
            amin and sizemults is given, then the larger multiplier for the two
            will be used.
        bmin : float, optional
            A minimum thickness to use for the b box vector direction of the
            final system.  Default value is 0.0.  For the non-periodic
            directions, the resulting vector multiplier will be even.  If both
            bmin and sizemults is given, then the larger multiplier for the two
            will be used.
        cmin : float, optional
            A minimum thickness to use for the c box vector direction of the
            final system.  Default value is 0.0.  For the non-periodic
            directions, the resulting vector multiplier will be even.  If both
            cmin and sizemults is given, then the larger multiplier for the two
            will be used.
        shift : float, optional
            A rigid body shift to apply to the rotated cell prior to inserting
            the dislocation.  Should be selected such that the ideal slip plane
            does not correspond to any atomic planes.  Is taken as absolute if
            shiftscale is False, or relative to the rotated cell's box vectors
            if shiftscale is True.  Cannot be given with shiftindex.  If
            neither shift nor shiftindex is given will use the shift set during
            class initialization.
        shiftindex : float, optional
            The index of the identified optimum shifts based on the rotated
            cell to use.  Different values allow for the selection of different
            atomic planes neighboring the slip plane.  Note that shiftindex
            values only apply shifts normal to the slip plane; best shifts for
            non-planar dislocations (like bcc screw) may also need a shift in
            the slip plane.  Cannot be given with shiftindex.  If neither shift
            nor shiftindex is given then shiftindex = 0 is used then will use
            the shift set during class initialization.
        shiftscale : bool, optional
            If False (default), a given shift value will be taken as absolute
            Cartesian.  If True, a given shift will be taken relative to the
            rotated cell's box vectors.
        boundaryshape : str, optional
            Indicates the shape of the boundary region to use.  Options are
            'cylinder' (default) and 'box'.  For 'cylinder', the non-boundary
            region is defined by a cylinder with axis along the dislocation
            line and a radius that ensures the boundary is at least
            boundarywidth thick.  For 'box', the boundary region will be
            exactly boundarywidth thick all around.      
        boundarywidth : float, optional
            The width of the boundary region to apply.  Default value is 0.0,
            i.e. no boundary region.  All atoms in the boundary region will
            have their atype values changed.
        boundaryscale : bool, optional
            If False (Default), the boundarywidth will be taken as absolute.
            If True, the boundarywidth will be taken relative to the magnitude
            of the unit cell's a box vector.
        return_base_system : bool, optional
            If True then the dislocation-free base system corresponding to the
            dislocation system will also be returned.  The base system is used
            as a reference state for most of the dislocation analysis tools.
        
        Returns
        -------
        base_system : atomman.System
            The base "perfect crystal" reference system associated with the
            dislocation system. Only returned if return_base_system is True.
        disl_system : atomman.System
            The generated dislocation monopole system.
        """
        # Set default sizemults
        if sizemults is None:
            sizemults = [2,2,2]
            sizemults[self.lineindex] = 1
        else:
            try:
                assert len(sizemults) == 3
                assert isinstance(sizemults[0], int) and sizemults[0] > 0
                assert isinstance(sizemults[1], int) and sizemults[1] > 0
                assert isinstance(sizemults[2], int) and sizemults[2] > 0
                assert sizemults[self.lineindex - 1] % 2 == 0
                assert sizemults[self.lineindex - 2] % 2 == 0
            except:
                raise TypeError('Invalid sizemults: must be 3 positive integers, and the two not along the dislocation line must be even')
        
        # Adjust multipliers based on min parameters
        if amin > 0.0:
            amult = int(np.ceil(amin / self.rcell.box.a))
            if self.lineindex != 0 and amult % 2 == 1:
                amult += 1
            if amult > sizemults[0]:
                sizemults[0] = amult

        if bmin > 0.0:
            bmult = int(np.ceil(bmin / self.rcell.box.b))
            if self.lineindex != 1 and bmult % 2 == 1:
                bmult += 1
            if bmult > sizemults[1]:
                sizemults[1] = bmult
                
        if cmin > 0.0:
            cmult = int(np.ceil(cmin / self.rcell.box.c))
            if self.lineindex != 2 and cmult % 2 == 1:
                cmult += 1
            if cmult > sizemults[2]:
                sizemults[2] = cmult
                
        # Modify the non-periodic size multipliers
        sizemults[self.lineindex] = (0, sizemults[self.lineindex])
        sizemults[self.lineindex - 1] = (-sizemults[self.lineindex - 1] // 2,
                                          sizemults[self.lineindex - 1] // 2)
        sizemults[self.lineindex - 2] = (-sizemults[self.lineindex - 2] // 2,
                                          sizemults[self.lineindex - 2] // 2)
        
        # Handle shift parameters
        if shift is not None:
            if shiftindex is not None:
                raise ValueError('shift and shiftindex cannot both be given')
            if shiftscale is True:
                self.__shift = miller.vector_crystal_to_cartesian(shift, self.rcell.box)
            else:
                self.__shift = np.asarray(shift)
                assert shift.shape == (3,) 
        elif shiftindex is not None:
            self.__shift = self.shifts[shiftindex]
        shift = self.shift
        
        # Handle boundary parameters
        if boundaryscale is True:
            boundarywidth = boundarywidth * self.ucell.box.a
        if boundaryshape not in ['cylinder', 'box']:
            raise ValueError('boundaryshape must be "cylinder" or "box"')
        
        # Create the system where the dislocation will be inserted
        base_system = self.rcell.supersize(*sizemults)
        base_system.atoms.pos += shift
        base_system.wrap()
        self.__base_system = base_system
        
        # Copy the system and displace atoms according to the dislocation solution
        disl_system = deepcopy(base_system)
        disl_system.atoms.pos += self.dislsol.displacement(disl_system.atoms.pos)
        disl_system.pbc = [False, False, False]
        disl_system.pbc[self.lineindex] = True
        disl_system.wrap()
        self.__disl_system = disl_system
        
        # Apply boundary region
        if boundarywidth > 0.0:
            
            if boundaryshape == 'box':
                shape = self.box_boundary(base_system.box, boundarywidth)
                
            elif boundaryshape == 'cylinder':
                shape = self.cylinder_boundary(base_system.box, boundarywidth)
            
            # Change atypes of atoms outside box
            disl_system.atoms.atype[shape.outside(disl_system.atoms.pos)] += base_system.natypes
            disl_system.symbols = 2 * base_system.symbols
        
        if return_base_system:
            return base_system, disl_system
        else:
            return disl_system

    def periodicarray(self, sizemults=None, amin=0.0, bmin=0.0, cmin=0.0,
                      shift=None, shiftindex=None, shiftscale=False,
                      boundarywidth=0.0, boundaryscale=False, linear=False,
                      cutoff=None, return_base_system=False):
        """
        Constructs a dislocation monopole atomic configuration containing a
        single perfectly straight dislocation. The resulting system will be
        periodic along the box vector direction that corresponds to the
        dislocation's line direction, and non-periodic in the other two box
        vector directions.  Boundary atoms near the two free surfaces will be
        identified by changing their atype values making it easy to identify
        them later for assigning different boundary conditions.
        
        Parameters
        ----------
        sizemults : tuple, optional
            The size multipliers to use when generating the system.  Values
            are limited to being positive integers.  The multipliers for the
            two non-periodic directions must be even.  If not given, the
            default multipliers will be 2 for the non-periodic directions and
            1 for the periodic direction.
        amin : float, optional
            A minimum thickness to use for the a box vector direction of the
            final system.  Default value is 0.0.  For the non-periodic
            directions, the resulting vector multiplier will be even.  If both
            amin and sizemults is given, then the larger multiplier for the two
            will be used.
        bmin : float, optional
            A minimum thickness to use for the b box vector direction of the
            final system.  Default value is 0.0.  For the non-periodic
            directions, the resulting vector multiplier will be even.  If both
            bmin and sizemults is given, then the larger multiplier for the two
            will be used.
        cmin : float, optional
            A minimum thickness to use for the c box vector direction of the
            final system.  Default value is 0.0.  For the non-periodic
            directions, the resulting vector multiplier will be even.  If both
            cmin and sizemults is given, then the larger multiplier for the two
            will be used.
        shift : float, optional
            A rigid body shift to apply to the rotated cell prior to inserting
            the dislocation.  Should be selected such that the ideal slip plane
            does not correspond to any atomic planes.  Is taken as absolute if
            shiftscale is False, or relative to the rotated cell's box vectors
            if shiftscale is True.  Cannot be given with shiftindex.  If
            neither shift nor shiftindex is given then shiftindex = 0 is used.
        shiftindex : float, optional
            The index of the identified optimum shifts based on the rotated
            cell to use.  Different values allow for the selection of different
            atomic planes neighboring the slip plane.  Note that shiftindex
            values only apply shifts normal to the slip plane; best shifts for
            non-planar dislocations (like bcc screw) may also need a shift in
            the slip plane.  Cannot be given with shiftindex.  If neither shift
            nor shiftindex is given then shiftindex = 0 is used.
        shiftscale : bool, optional
            If False (default), a given shift value will be taken as absolute
            Cartesian.  If True, a given shift will be taken relative to the
            rotated cell's box vectors.
        boundarywidth : float, optional
            The width of the boundary region to apply.  Default value is 0.0,
            i.e. no boundary region.  All atoms in the boundary region will
            have their atype values changed and will be displaced by linear
            displacements.
        boundaryscale : bool, optional
            If False (Default), the boundarywidth will be taken as absolute.
            If True, the boundarywidth will be taken relative to the magnitude
            of the unit cell's a box vector.
        linear : bool, optional
            If True, then only linear displacements will be used and not the
            dislocation solution.  Using only linear displacements is useful
            for screw dislocations and dislocations with large stacking fault
            distances.  If False (default) then the dislocation solution will
            be used for the middle displacements and linear displacements only
            in the boundary region.
        cutoff : float, optional
            Cutoff distance to use for identifying duplicate atoms to remove.
            For dislocations with an edge component, applying the displacements
            creates an extra half-plane of atoms that will have (nearly)
            identical positions with other atoms after altering the boundary
            conditions.  Default value is 0.5 Angstrom.
        return_base_system : bool, optional
            If True then the dislocation-free base system corresponding to the
            dislocation system will also be returned.  The base system is used
            as a reference state for most of the dislocation analysis tools.
        
        Returns
        -------
        base_system : atomman.System
            The base "perfect crystal" reference system associated with the
            dislocation system.  If the Burgers vector has an edge component
            then the atoms deleted when generating disl_system will also be 
            deleted from base_system. Only returned if return_base_system is
            True.
        disl_system : atomman.System
            The generated periodic array of dislocations system.
        """
        # Set default sizemults
        if sizemults is None:
            sizemults = [2,2,2]
            sizemults[self.lineindex] = 1
        else:
            try:
                assert len(sizemults) == 3
                assert isinstance(sizemults[0], int) and sizemults[0] > 0
                assert isinstance(sizemults[1], int) and sizemults[1] > 0
                assert isinstance(sizemults[2], int) and sizemults[2] > 0
                assert sizemults[self.lineindex - 1] % 2 == 0
                assert sizemults[self.lineindex - 2] % 2 == 0
            except:
                raise TypeError('Invalid sizemults: must be 3 positive integers, and the two not along the dislocation line must be even')
        
        # Adjust multipliers based on min parameters
        if amin > 0.0:
            amult = int(np.ceil(amin / self.rcell.box.a))
            if self.lineindex != 0 and amult % 2 == 1:
                amult += 1
            if amult > sizemults[0]:
                sizemults[0] = amult

        if bmin > 0.0:
            bmult = int(np.ceil(bmin / self.rcell.box.b))
            if self.lineindex != 1 and bmult % 2 == 1:
                bmult += 1
            if bmult > sizemults[1]:
                sizemults[1] = bmult
                
        if cmin > 0.0:
            cmult = int(np.ceil(cmin / self.rcell.box.c))
            if self.lineindex != 2 and cmult % 2 == 1:
                cmult += 1
            if cmult > sizemults[2]:
                sizemults[2] = cmult
                
        # Modify the non-periodic size multipliers
        sizemults[self.lineindex] = (0, sizemults[self.lineindex])
        sizemults[self.lineindex - 1] = (-sizemults[self.lineindex - 1] // 2,
                                          sizemults[self.lineindex - 1] // 2)
        sizemults[self.lineindex - 2] = (-sizemults[self.lineindex - 2] // 2,
                                          sizemults[self.lineindex - 2] // 2)
        
        # Handle shift parameters
        if shift is not None:
            if shiftindex is not None:
                raise ValueError('shift and shiftindex cannot both be given')
            if shiftscale is True:
                self.__shift = miller.vector_crystal_to_cartesian(shift, self.rcell.box)
            else:
                self.__shift = np.asarray(shift)
                assert shift.shape == (3,) 
        elif shiftindex is not None:
            self.__shift = self.shifts[shiftindex]
        shift = self.shift
        
        # Handle boundary parameters
        if boundaryscale is True:
            boundarywidth = boundarywidth * self.ucell.box.a
            
        # Create the system where the dislocation will be inserted
        base_system = self.rcell.supersize(*sizemults)
        base_system.atoms.pos += shift
        base_system.wrap()
        
        # Create the dislocation system
        if linear is True:
            disl_system = dislocation_array(base_system, m=self.dislsol.m, n=self.dislsol.n,
                                            burgers=self.dislsol.burgers, cutoff=cutoff)
        else:
            disl_system = dislocation_array(base_system, self.dislsol,
                                            bwidth=boundarywidth, cutoff=cutoff)
        self.__disl_system = disl_system
        base_system = base_system.atoms_ix[disl_system.atoms.old_id]
        self.__base_system = base_system

        # Apply boundary region
        if boundarywidth > 0.0:
            
            shape = self.array_boundary(base_system.box, boundarywidth)
            
            # Change atypes of atoms outside box
            disl_system.atoms.atype[shape.outside(disl_system.atoms.pos)] += base_system.natypes
            disl_system.symbols = 2 * base_system.symbols
        
        if return_base_system:
            return base_system, disl_system
        else:
            return disl_system
        


# coding: utf-8

class Shape():
    """
    Template class for defining geometric regions in space.
    """

    def __init__(self):
        """
        Defines a shape
        """
        raise NotImplementedError('Base shape class cannot be directly used')
        
    def inside(self, pos, inclusive=True):
        """
        Indicates if position(s) are inside the shape.
        
        Parameters
        ----------
        pos : array-like object
            Nx3 array of coordinates. 
        inclusive : bool, optional
            Indicates if points on the shape's boundaries are to be included.
            Default value is True.
        
        Returns
        -------
        numpy.NDArray
            N array of bool values: True if inside shape
        """
        raise NotImplementedError('Base shape class cannot be directly used')
        
    def outside(self, pos, inclusive=False):
        """
        Indicates if position(s) are inside the shape.
        
        Parameters
        ----------
        pos : array-like object
            Nx3 array of coordinates. 
        inclusive : bool, optional
            Indicates if points on the shape's boundaries are to be included.
            Default value is False.
        
        Returns
        -------
        numpy.NDArray
            N array of bool values: True if outside shape
        """
        return ~self.inside(pos, inclusive=not inclusive)
        


# coding: utf-8

# http://www.numpy.org/
import numpy as np

def vect_angle(vect1, vect2, unit='degree'):
    """
    Returns the angle between two vectors.
    
    Parameters
    ----------
    vect1 : list or numpy.ndarray
        First vector.
    vect2 : list or numpy.ndarray
        Second vector.
    unit : str
        Specifies unit of returned angle: 'degree' or 'radian'. Default value
        is 'degree'.
        
    Returns
    -------
    float
        The angle between vect1 and vect2.
    """
    # Convert to numpy arrays if needed
    vect1 = np.asarray(vect1)
    vect2 = np.asarray(vect2)
    
    # Compute unit vectors
    u_vect1 = vect1 / np.linalg.norm(vect1)
    u_vect2 = vect2 / np.linalg.norm(vect2)
    
    # calculate cosine angle between vectors
    cosine = np.dot(u_vect1, u_vect2)
    
    # Clean up any invalid numbers due to rounding
    try:
        # For multiple values
        cosine[cosine < -1] = -1
        cosine[cosine > 1] = 1
    except:
        # For single values
        if cosine < -1:
            cosine = -1
        elif cosine > 1:
            cosine = 1
    
    # Convert to degrees and return
    if unit == 'degree':
        return 180 * np.arccos(cosine) / np.pi
    elif unit == 'radian':
        return np.arccos(cosine)
    else:
        raise ValueError("unit must be 'degree' or 'radian'.")


# coding: utf-8
# cython: language_level=3
# Standard Python libraries
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)
from copy import deepcopy

# http://cython.org/
import cython

# http://www.numpy.org/
import numpy as np

@cython.boundscheck(False)
@cython.wraparound(False)
def dvect(pos_0, pos_1, box, pbc):
    """
    Computes the shortest vector between pos_0 and pos_1 using box
    dimensions and accounting for periodic boundaries.

    Parameters
    ----------
    pos_0 : numpy.ndarray, list, or tuple
        Absolute Cartesian vector position(s) to use as reference point(s).
    pos_1 : numpy.ndarray, list, or tuple
        Absolute Cartesian vector position(s) to find relative to pos_0.
    box : atomman.Box
        Defines the system/box dimensions
    pbc : list, tuple, or numpy.ndarray of bool.
        Three Boolean values indicating which of the three box vectors are
        periodic (True means periodic).

    Returns
    -------
    numpy.ndarray
        The shortest vectors from each pos_0 to pos_1 positions.
    """
    
    # Convert pos_0 to numpy array with proper dimensions
    pos_0 = np.asarray(pos_0, dtype=np.float64)
    if pos_0.ndim == 0:
        raise TypeError('Invalid pos_0')
    if pos_0.ndim == 1:
        pos_0 = pos_0[np.newaxis, :]

    # Convert pos_1 to numpy array with proper dimensions
    pos_1 = np.asarray(pos_1, dtype=np.float64)
    if pos_1.ndim == 0:
        raise TypeError('Invalid pos_1')
    if pos_1.ndim == 1:
        pos_1 = pos_1[np.newaxis, :]

    # Broadcast to compatible lengths
    if len(pos_0) == 1:
        pos_0 = np.broadcast_to(pos_0, pos_1.shape)
    elif len(pos_1) == 1:
        pos_1 = np.broadcast_to(pos_1, pos_0.shape)
    elif len(pos_0) != len(pos_1):
        raise ValueError('Incompatible pos lengths')

    # Extract box vectors
    bvects = box.vects

    # Call the cython function
    return dvect_c(pos_0, pos_1, bvects, pbc[0], pbc[1], pbc[2])

@cython.boundscheck(False)
@cython.wraparound(False)
cdef dvect_c(const double[:,:] pos_0,
             const double[:,:] pos_1, 
             const double[:,:] bvects,
             const bint pbc_x,
             const bint pbc_y,
             const bint pbc_z):
    """
    Computes the shortest distance between pos_0 and pos_1 using box 
    dimensions and accounting for periodic boundaries.
    
    Parameters
    ----------
    pos_0 : cython.memoryview
        Absolute Cartesian vector position(s) to use as reference point(s).
    pos_1 : cython.memoryview
        Absolute Cartesian vector position(s) to find relative to pos_0.
    bvects : cython.memoryview
        3x3 array defining the system/box dimensions.
    pbc_x : bint
        Flag indicating to make x periodic.
    pbc_y : bint
        Flag indicating to make y periodic.
    pbc_z : bint
        Flag indicating to make z periodic.

    Returns
    -------
    cython.memoryview
        The shortest vectors from each pos_0 to pos_1 positions.
    """
    
    # Define parameters
    cdef Py_ssize_t ni = pos_0.shape[0]
    cdef int nj = 3
    cdef int i, j, x, y, z, xl, xh, yl, yh, zl, zh    
    cdef double[:] test = np.empty(3)
    cdef double mag_test, mag_d
    
    # Define output array and its view
    d = np.empty_like(pos_0, np.float64)
    cdef double[:,:] dv = d
    
    # Create iterators based on pbc
    if pbc_x:
        xl, xh = -1, 2
    else:
        xl, xh = 0, 1
    if pbc_y:
        yl, yh = -1, 2
    else:
        yl, yh = 0, 1
    if pbc_z:
        zl, zh = -1, 2
    else:
        zl, zh = 0, 1
    
    # Loop over all pos
    for i in range(ni):
       
        # Compute pos_1 - pos_0 
        for j in range(nj):
            dv[i,j] = pos_1[i,j] - pos_0[i,j]
        
        # Loop over all periodic boundary conditions
        for x in range(xl, xh):
            for y in range(yl, yh):
                for z in range(zl, zh):
                    if x == 0 and y == 0 and z == 0:
                        continue
                    
                    # Compute pos_1 - pos_0 + boundary image shifts
                    for j in range(nj):
                        test[j] = (pos_1[i,j] - pos_0[i,j] 
                                   + x * bvects[0,j] 
                                   + y * bvects[1,j] 
                                   + z * bvects[2,j])
                    
                    # Replace d if new vector is smaller
                    mag_test = test[0] * test[0] + test[1] * test[1] + test[2] * test[2]
                    mag_d = dv[i,0] * dv[i,0] + dv[i,1] * dv[i,1] + dv[i,2] * dv[i,2]
                    if mag_test < mag_d:
                        for j in range(nj):
                            dv[i,j] = test[j]

    return d


# coding: utf-8
def screen_input(prompt=''):
    """
    Replacement input function that is compatible with the mingw terminal.
    
    Parameters
    ----------
    prompt : str, optional
        The screen prompt to use for asking for the input.
        
    Returns
    -------
    str
        The user input.
    """
    # Flush prompt to work interactively with mingw
    print(prompt, end=' ', flush=True)
    return input()


# coding: utf-8

# http://www.numpy.org/
import numpy as np

from .. import displacement

def disregistry(basesystem, dislsystem, m=[1.0, 0.0, 0.0],
                n=[0.0, 1.0, 0.0], planepos=[0.0, 0.0, 0.0]):
    """
    Computes the disregistry profile for a dislocation system.
    
    Parameters
    ----------
    basesystem : atomman.System
        A perfect reference system with atoms directly corresponding to atoms
        in dislsystem.
    dislsystem : atomman.System
        A dislocation-containing system.
    m : array-like object, optional
        The dislocation solution m unit vector.  This vector is in the slip
        plane and perpendicular to the dislocation line direction.  Default
        value is [1,0,0] (Cartesian x-axis).
    n : array-like object, optional
        The dislocation solution n unit vector.  This vector is normal to the 
        slip plane.  Only needed if dislsol is not given.  Default
        value is [0,1,0] (Cartesian y-axis).
    planepos : array-like object, optional
        A position on the slip plane so that the plane can be fully defined.
        The slip plane position should fall between two planes of atoms.
        Default value is [0,0,0].
    
    Returns
    -------
    coord : numpy.ndarray
        The (N,) array of unique coord-coordinates (atomic columns) neighboring
        the slip plane.
    disregistry : numpy.ndarray
        A (N, 3) array of the dislocation's disregistry at each coord.
    """
    # Handle m, n and planepos
    m = np.asarray(m, dtype=float)
    n = np.asarray(n, dtype=float)
    planepos = np.asarray(planepos, dtype=float)
    #if not np.isclose(m.dot(planepos), 0.0, atol=1e-8, rtol=0.0):
    #    raise ValueError('m and planepos must be perpendicular')
    
    # Extract pos from basesystem and compute atomic displacements
    basepos = basesystem.atoms.pos
    disp = displacement(basesystem, dislsystem)
    
    # Transform basepos and planepos to calculation coordinates
    allx = np.dot(basepos, m)
    ally = np.dot(basepos, n)
    midy = np.dot(planepos, n)
    
    # Identify atoms just above and below slip plane
    uniquey = np.unique(ally)
    abovey = uniquey[uniquey > midy].min()
    belowy = uniquey[uniquey < midy].max()
    if np.isclose(abovey, belowy):
        raise ValueError('planepos must fall between atomic planes')
    #uniquey = np.unique(basepos[:, 1])
    #abovey = uniquey[uniquey > 0].min()
    #belowy = uniquey[uniquey < 0].max()
    
    # Get coordinates of atoms just above and below slip plane
    abovex = allx[np.isclose(ally, abovey)]
    belowx = allx[np.isclose(ally, belowy)]
    #abovex = basepos[np.isclose(basepos[:, 1], abovey)][:, 0]
    #belowx = basepos[np.isclose(basepos[:, 1], belowy)][:, 0]
    
    # Identify unique coordinates
    uabovex = np.unique(abovex)
    ubelowx = np.unique(belowx)
    coord = np.union1d(uabovex, ubelowx)
    
    # Get displacements of atoms just above and below slip plane
    abovedisp = disp[np.isclose(ally, abovey)]
    belowdisp = disp[np.isclose(ally, belowy)]
    #abovedisp = disp[np.isclose(basepos[:, 1], abovey)]
    #belowdisp = disp[np.isclose(basepos[:, 1], belowy)]
    
    # Average displacements for the same coordinates
    abovedispmean = np.empty((len(uabovex), 3))
    for i, ix in enumerate(uabovex):
        abovedispmean[i] = abovedisp[np.isclose(abovex, ix)].mean(axis=0)
    belowdispmean = np.empty((len(ubelowx), 3))
    for i, ix in enumerate(ubelowx):
        belowdispmean[i] = belowdisp[np.isclose(belowx, ix)].mean(axis=0)
    
    # Linearly interpolate displacement values to all coord
    abovedispinterp = np.vstack([np.interp(coord, uabovex, abovedispmean[:,0]),
                                 np.interp(coord, uabovex, abovedispmean[:,1]),
                                 np.interp(coord, uabovex, abovedispmean[:,2])]).T
    belowdispinterp = np.vstack([np.interp(coord, ubelowx, belowdispmean[:,0]),
                                 np.interp(coord, ubelowx, belowdispmean[:,1]),
                                 np.interp(coord, ubelowx, belowdispmean[:,2])]).T
    
    # Calculate disregistry
    disregistry = abovedispinterp - belowdispinterp
    
    return coord, disregistry


# coding: utf-8

from .ase_Atoms import dump as dump_ase_Atoms
from .pymatgen_Structure import dump as dump_pymatgen_Structure
from .table import dump as dump_table
from .system_model import dump as dump_system_model
from .poscar import dump as dump_poscar
from .atom_data import dump as dump_atom_data
from .atom_dump import dump as dump_atom_dump
from .spglib_cell import dump as dump_spglib_cell
from .phonopy_Atoms import dump as dump_phonopy_Atoms
from .lammps_commands import dump as dump_lammps_commands

__all__ = ['dump', 'dump_ase_Atoms', 'dump_pymatgen_Structure', 'dump_table',
           'dump_system_model', 'dump_poscar', 'dump_atom_data',
           'dump_atom_dump', 'dump_spglib_cell', 'dump_phonopy_Atoms',
           'dump_lammps_commands']

def dump(style, system, **kwargs):
    """
    Convert a System to another format.
    
    Parameters
    ----------
    style : str
        Indicates the format of the content to dump the atomman.System as.
    system : atomman.System
        The system to convert.
    kwargs : any, optional
        Any extra keyword arguments to pass to the underlying dump methods.
        
    Returns
    -------
    str, object or tuple
        Any content returned by the underlying dump methods.
    """
    
    if style == 'system_model':
        return dump_system_model(system, **kwargs)
    
    elif style == 'atom_data':
        return dump_atom_data(system, **kwargs)
    
    elif style == 'atom_dump':
        return dump_atom_dump(system, **kwargs)
    
    elif style == 'table':
        return dump_table(system, **kwargs)
    
    elif style == 'ase_Atoms':
        return dump_ase_Atoms(system, **kwargs)

    elif style == 'phonopy_Atoms':
        return dump_phonopy_Atoms(system, **kwargs)
    
    elif style == 'pymatgen_Structure':
        return dump_pymatgen_Structure(system, **kwargs)
    
    elif style == 'poscar':
        return dump_poscar(system, **kwargs)
    
    elif style == 'spglib_cell':
        return dump_spglib_cell(system, **kwargs)

    elif style == 'lammps_commands':
        return dump_lammps_commands(system, **kwargs)
        
    else:
        raise ValueError('Unsupported style')


# coding: utf-8
# Standard Python libraries
from copy import deepcopy

# http://www.numpy.org/
import numpy as np

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# atomman imports
import atomman.unitconvert as uc
from ..tools import vect_angle
from ..region import Shape, Plane

class Box(Shape, object):
    """
    A representation of a triclinic (parallelepiped) box.
    """
    
    def __init__(self, **kwargs):
        """
        Initializes a Box.  If parameters besides origin are given they
        must completely define the box.  Allowed parameter sets are:
        
        - no parameters -> box is set to square unit box with origin = [0,0,0].
        
        - origin. -> Only origin is changed (same as setting origin directly).
        
        - vects, (and origin).
        
        - avect, bvect, cvect, (and origin).
        
        - a, b, c, (alpha, beta, gamma, and origin).
        
        - lx, ly, lz, (xy, xz, yz, and origin).
        
        - xlo, xhi, ylo, yhi, zlo, zhi, (xy, xz, and yz).

        - model
        
        See the description of class methods and attributes for more details
        on the allowed parameters.
        
        """
        self.__vects = np.eye(3, dtype='float64')
        self.__origin = np.zeros(3, dtype='float64')
        
        if len(kwargs) > 0:
            if 'model' in kwargs:
                if len(kwargs) > 1:
                    raise ValueError('model cannot be given with other parameters')
                self.model(kwargs['model'])
            else:
                self.set(**kwargs)
    
    @classmethod
    def cubic(cls, a):
        """
        Initializes a Box in standard cubic setting using only cubic lattice
        parameters.  

        a = b = c, alpha = beta = gamma = 90

        Parameters
        ----------
        - a : float
            The a lattice constant

        Returns
        -------
        atomman.Box
        """
        return cls(a=a, b=a, c=a, alpha=90, beta=90, gamma=90)

    @classmethod
    def hexagonal(cls, a, c):
        """
        Initializes a Box in standard hexagonal setting using only hexagonal lattice
        parameters.  

        a = b != c, alpha = beta = 90, gamma = 120

        Parameters
        ----------
        - a : float
            The a lattice constant
        - c : float
            The c lattice constant

        Returns
        -------
        atomman.Box
        """
        if a == c:
            raise ValueError('hexagonal lattice constants must be different')

        return cls(a=a, b=a, c=c, alpha=90, beta=90, gamma=120)

    @classmethod
    def tetragonal(cls, a, c):
        """
        Initializes a Box in standard tetragonal setting using only tetragonal lattice
        parameters.  

        a = b != c, alpha = beta = gamma = 90

        Parameters
        ----------
        - a : float
            The a lattice constant
        - c : float
            The c lattice constant

        Returns
        -------
        atomman.Box
        """
        if a == c:
            raise ValueError('tetragonal lattice constants must be different')

        return cls(a=a, b=a, c=c, alpha=90, beta=90, gamma=90)

    @classmethod
    def trigonal(cls, a, alpha):
        """
        Initializes a Box in standard trigonal setting using only trigonal lattice
        parameters.  

        a = b = c, alpha = beta = gamma < 120

        Parameters
        ----------
        - a : float
            The a lattice constant
        - alpha : float
            The alpha lattice angle in degrees

        Returns
        -------
        atomman.Box
        """
        if alpha >= 120.0:
            raise ValueError('trigonal alpha angle must be less than 120 degrees')        

        return cls(a=a, b=a, c=a, alpha=alpha, beta=alpha, gamma=alpha)

    @classmethod
    def orthorhombic(cls, a, b, c):
        """
        Initializes a Box in standard orthorhombic setting using only orthorhombic lattice
        parameters.  

        a != b != c, alpha = beta = gamma = 90

        Parameters
        ----------
        - a : float
            The a lattice constant
        - b : float
            The b lattice constant
        - c : float
            The c lattice constant

        Returns
        -------
        atomman.Box
        """
        if a == b or a == c:
            raise ValueError('orthorhombic lattice constants must be different')

        return cls(a=a, b=b, c=c, alpha=90, beta=90, gamma=90)

    @classmethod
    def monoclinic(cls, a, b, c, beta):
        """
        Initializes a Box in standard monoclinic setting using only monoclinic lattice
        parameters.  

        a != b != c, alpha = gamma = 90, beta > 90

        Parameters
        ----------
        - a : float
            The a lattice constant
        - b : float
            The b lattice constant
        - c : float
            The c lattice constant
        - beta : float
            The beta lattice angle in degrees

        Returns
        -------
        atomman.Box
        """
        if a == b or a == c:
            raise ValueError('monoclinic lattice constants must be different')
        if beta <= 90.0:
            raise ValueError('monoclinic angle beta must be greater than 90 degrees')

        return cls(a=a, b=b, c=c, alpha=90, beta=beta, gamma=90)

    @classmethod
    def triclinic(cls, a, b, c, alpha, beta, gamma):
        """
        Initializes a Box in standard triclinic setting using only triclinic lattice
        parameters.  

        a != b != c, alpha != beta != gamma 

        Parameters
        ----------
        - a : float
            The a lattice constant
        - b : float
            The b lattice constant
        - c : float
            The c lattice constant
        - alpha : float
            The alpha lattice angle in degrees
        - beta : float
            The beta lattice angle in degrees
        - gamma : float
            The gamma lattice angle in degrees

        Returns
        -------
        atomman.Box
        """
        if a == b or a == c:
            raise ValueError('monoclinic lattice constants must be different')
        if alpha == beta or alpha == gamma:
            raise ValueError('monoclinic lattice angles must be different')

        return cls(a=a, b=b, c=c, alpha=alpha, beta=beta, gamma=gamma)

    @property
    def vects(self):
        """numpy.ndarray : Array containing all three box vectors.  Can be set directly."""
        return deepcopy(self.__vects)
    
    @vects.setter
    def vects(self, value):
        self.__vects[:] = value
        
        #Zero out near zero terms
        self.__vects[np.isclose(self.__vects/abs(self.__vects).max(), 0.0, atol=1e-9)] = 0.0
    
    @property
    def origin(self):
        """numpy.ndarray : Box origin position where vects are added to define the box.  Can be set directly."""
        return deepcopy(self.__origin)
    
    @origin.setter
    def origin(self, value):
        self.__origin[:] = value
    
    @property
    def avect(self):
        """numpy.ndarray : Vector associated with the a box dimension."""
        return self.vects[0]
    
    @property
    def bvect(self):
        """numpy.ndarray : Vector associated with the b box dimension."""
        return self.vects[1]
    
    @property
    def cvect(self):
        """numpy.ndarray : Vector associated with the c box dimension."""
        return self.vects[2]
    
    @property
    def a(self):
        """float : The a lattice parameter (magnitude of avect)."""
        return (self.__vects[0,0]**2 + self.__vects[0,1]**2 + self.__vects[0,2]**2)**0.5
    
    @property
    def b(self):
        """float : The b lattice parameter (magnitude of avect)."""
        return (self.__vects[1,0]**2 + self.__vects[1,1]**2 + self.__vects[1,2]**2)**0.5
    
    @property
    def c(self):
        """float : The c lattice parameter (magnitude of avect)."""
        return (self.__vects[2,0]**2 + self.__vects[2,1]**2 + self.__vects[2,2]**2)**0.5 
    
    @property
    def alpha(self):
        """float : The alpha lattice angle in degrees (angle between bvect and cvect)."""
        return vect_angle(self.__vects[1], self.__vects[2])
    
    @property
    def beta(self):
        """float : The beta lattice angle in degrees (angle between avect and cvect)."""
        return vect_angle(self.__vects[0], self.__vects[2])
    
    @property
    def gamma(self):
        """float : The gamma lattice angle in degrees (angle between avect and bvect)."""
        return vect_angle(self.__vects[0], self.__vects[1])
    
    @property
    def lx(self):
        """float : LAMMPS lx box length (avect[0] for normalized boxes)."""
        assert self.is_lammps_norm(), 'Box is not normalized for LAMMPS style parameters'
        return self.__vects[0,0]
    
    @property
    def ly(self):
        """float : LAMMPS ly box length (bvect[1] for normalized boxes)."""
        assert self.is_lammps_norm(), 'Box is not normalized for LAMMPS style parameters'
        return self.__vects[1,1]
    
    @property
    def lz(self):
        """float : LAMMPS lz box length (cvect[2] for normalized boxes)."""
        assert self.is_lammps_norm(), 'Box is not normalized for LAMMPS style parameters'
        return self.__vects[2,2]
    
    @property
    def xy(self):
        """float : LAMMPS xy box tilt factor (bvect[0] for normalized boxes)."""
        assert self.is_lammps_norm(), 'Box is not normalized for LAMMPS style parameters'
        return self.__vects[1,0]
    
    @property
    def xz(self):
        """float : LAMMPS xz box tilt factor (cvect[0] for normalized boxes)."""
        assert self.is_lammps_norm(), 'Box is not normalized for LAMMPS style parameters'
        return self.__vects[2,0]
    
    @property
    def yz(self):
        """float : LAMMPS yz box tilt factor (cvect[1] for normalized boxes)."""
        assert self.is_lammps_norm(), 'Box is not normalized for LAMMPS style parameters'
        return self.__vects[2,1]
    
    @property
    def xlo(self):
        """float : LAMMPS xlo box lo term (origin[0] for normalized boxes)."""
        assert self.is_lammps_norm(), 'Box is not normalized for LAMMPS style parameters'
        return self.__origin[0]
    
    @property
    def ylo(self):
        """float : LAMMPS ylo box lo term (origin[1] for normalized boxes)."""
        assert self.is_lammps_norm(), 'Box is not normalized for LAMMPS style parameters'
        return self.__origin[1]
    
    @property
    def zlo(self):
        """float : LAMMPS zlo box lo term (origin[2] for normalized boxes)."""
        assert self.is_lammps_norm(), 'Box is not normalized for LAMMPS style parameters'
        return self.__origin[2]
    
    @property
    def xhi(self):
        """float : LAMMPS xhi box hi term (origin[0] + lx for normalized boxes)."""
        assert self.is_lammps_norm(), 'Box is not normalized for LAMMPS style parameters'
        return self.__origin[0] + self.__vects[0,0]
    
    @property
    def yhi(self):
        """float : LAMMPS yhi box hi term (origin[1] + ly for normalized boxes)."""
        assert self.is_lammps_norm(), 'Box is not normalized for LAMMPS style parameters'
        return self.__origin[1] + self.__vects[1,1]
    
    @property
    def zhi(self):
        """float : LAMMPS zhi box hi term (origin[2] + lz for normalized boxes)."""
        assert self.is_lammps_norm(), 'Box is not normalized for LAMMPS style parameters'
        return self.__origin[2] + self.__vects[2,2]
    
    @property
    def volume(self):
        """float : The volume of the box."""
        return np.abs(np.dot(self.avect, np.cross(self.bvect, self.cvect)))

    @property
    def planes(self):
        """tuple : The box's planes represented as atomman.region.Plane objects."""
        return (Plane(np.cross(self.cvect, self.bvect), self.origin),
                Plane(np.cross(self.avect, self.cvect), self.origin),
                Plane(np.cross(self.bvect, self.avect), self.origin),
                Plane(np.cross(self.bvect, self.cvect), self.origin + self.avect),
                Plane(np.cross(self.cvect, self.avect), self.origin + self.bvect),
                Plane(np.cross(self.avect, self.bvect), self.origin + self.cvect))

    def __str__(self):
        """
        The string representation of the box.  Lists the three vectors and origin.
        """
        return '\n'.join(['avect =  [%6.3f, %6.3f, %6.3f]' % (self.__vects[0,0], self.__vects[0,1], self.__vects[0,2]),
                          'bvect =  [%6.3f, %6.3f, %6.3f]' % (self.__vects[1,0], self.__vects[1,1], self.__vects[1,2]),
                          'cvect =  [%6.3f, %6.3f, %6.3f]' % (self.__vects[2,0], self.__vects[2,1], self.__vects[2,2]),
                          'origin = [%6.3f, %6.3f, %6.3f]' % (self.__origin[0],  self.__origin[1],  self.__origin[2])])
    
    def model(self, model=None, length_unit='angstrom'):
        """
        Reads or generates a data model for the box.

        Parameters
        ----------
        model : str or DataModelDict, optional
            JSON/XML formatted data, or path to file containing said data.  If
            not given, then a model for the current box will be returned.
        length_unit : str, optional
            Unit of length to save box values in if data model is to be
            generated.  Default value is 'angstrom'.

        Returns
        -------
        DataModelDict.DataModelDict
            A JSON/XML equivalent data model for the box.  Returned if model
            is not given
        """
        # Set values if model given
        if model is not None:
            
            # Find box element
            model = DM(model).find('box')
            avect = uc.value_unit(model['avect'])
            bvect = uc.value_unit(model['bvect'])
            cvect = uc.value_unit(model['cvect'])
            origin = uc.value_unit(model['origin'])
            self.set(avect=avect, bvect=bvect, cvect=cvect, origin=origin)
        
        # Return DataModelDict if model not given
        else:
            model = DM()
            model['box'] = DM()
            model['box']['avect'] = uc.model(self.avect, length_unit)
            model['box']['bvect'] = uc.model(self.bvect, length_unit)
            model['box']['cvect'] = uc.model(self.cvect, length_unit)
            model['box']['origin']= uc.model(self.origin, length_unit)

            return model

    def set(self, **kwargs):
        """
        Sets a Box's dimensions.  If parameters besides origin are given they
        must completely define the box.  Allowed parameter sets are:
        
        - no parameters -> box is set to square unit box with origin = [0,0,0].
        
        - origin. -> Only origin is changed (same as setting origin directly).
        
        - vects, (and origin).
        
        - avect, bvect, cvect, (and origin).
        
        - a, b, c, (alpha, beta, gamma, and origin).
        
        - lx, ly, lz, (xy, xz, yz, and origin).
        
        - xlo, xhi, ylo, yhi, zlo, zhi, (xy, xz, and yz).
        
        See the description of class methods and attributes for more details
        on the allowed parameters.
        """
        
        # Set default values if no kwargs given
        if len(kwargs) == 0:
            self.vects = np.eye(3)
            self.origin = np.zeros(3)
        
        # Set directly if vects given
        elif 'vects' in kwargs:
            vects = kwargs.pop('vects')
            origin = kwargs.pop('origin', [0.0, 0.0, 0.0])
            assert len(kwargs) == 0, 'Invalid arguments'
            self.vects = vects
            self.origin = origin
        
        # Call set_vectors if vect inputs given
        elif 'avect' in kwargs:
            self.set_vectors(**kwargs)
        
        # Call set_lengths if length inputs given
        elif 'lx' in kwargs:
            self.set_lengths(**kwargs)
        
        # Call set_hi_los if hi/lo inputs given
        elif 'xlo' in kwargs:
            self.set_hi_los(**kwargs)
        
        # Call set_abc if vector magnitudes are given
        elif 'a' in kwargs:
            self.set_abc(**kwargs)
        
        # Set only origin if given alone
        elif 'origin' in kwargs:
            origin = kwargs.pop('origin')
            assert len(kwargs) == 0, 'Invalid arguments'
            self.origin = origin
        
        else:
            raise TypeError('Invalid arguments')
    
    def set_vectors(self, avect, bvect, cvect, origin=None):
        """
        Set the box using the three box vectors.
        
        Parameters
        ----------
        avect : numpy.ndarray
            The 3D vector for the a box dimension.
        bvect : numpy.ndarray
            The 3D vector for the b box dimension.
        cvect : numpy.ndarray
            The 3D vector for the c box dimension.
        origin : numpy.ndarray, optional
            The 3D vector for the box origin position.  Default value is
            (0,0,0).
        """
        # Set default origin
        if origin is None:
            origin = [0.0, 0.0, 0.0]
        
        # Combine avect, bvect and cvect into vects and set directly
        self.vects = [avect, bvect, cvect]
        self.origin = origin
        
    def set_abc(self, a, b, c, alpha=90.0, beta=90.0, gamma=90.0, origin=None):
        """
        Set the box using crystal cell lattice parameters and angles.
        
        Parameters
        ----------
        a : float
            The a lattice parameter.
        b : float
            The b lattice parameter.
        c : float
            The c lattice parameter.
        alpha : float, optional
            The alpha lattice angle in degrees (angle between b and c vectors).
            Default value is 90.0.
        beta : float, optional
            The beta lattice angle in degrees (angle between a and c vectors).
            Default value is 90.0.
        gamma : float, optional
            The gamma lattice angle in degrees (angle between a and b vectors).
            Default value is 90.0.
        origin : numpy.ndarray, optional
            The 3D vector for the box origin position.  Default value is
            (0,0,0).
        """
        # Check that angles are between 0 and 180
        if alpha <= 0 or alpha >=180 or beta <= 0 or beta >= 180 or gamma <=0 or gamma >=180:
            raise ValueError('lattice angles must be between 0 and 180 degrees')

        # Convert to lx, ly, lz, xy, xz, yz
        lx = a
        xy = b * np.cos(gamma * np.pi / 180)
        xz = c * np.cos(beta * np.pi / 180)
        ly = (b**2 - xy**2)**0.5
        yz = (b * c * np.cos(alpha * np.pi / 180) - xy * xz) / ly
        lz = (c**2 - xz**2 - yz**2)**0.5

        # Call set_lengths
        self.set_lengths(lx=lx, ly=ly, lz=lz, xy=xy, xz=xz, yz=yz, origin=origin)
        
    def set_lengths(self, lx, ly, lz, xy=0.0, xz=0.0, yz=0.0, origin=None):
        """
        Set the box using LAMMPS box lengths and tilt factors.
        
        Parameters
        ----------
        lx : float
            The LAMMPS box length in the x direction.
        ly : float
            The LAMMPS box length in the y direction.
        lz : float
            The LAMMPS box length in the z direction.
        xy : float, optional
            The LAMMPS box tilt factor in the xy direction.  Default value is 
            0.0.
        xz : float, optional
            The LAMMPS box tilt factor in the xz direction.  Default value is 
            0.0.
        yz : float, optional
            The LAMMPS box tilt factor in the yz direction.  Default value is 
            0.0.
        origin : numpy.ndarray, optional
            The 3D vector for the box origin position.  Default value is
            (0,0,0).
        """
        
        assert lx > 0 and ly > 0 and lz > 0, 'box lengths must be positive'
        
        # Set default origin
        if origin is None:
            origin = [0.0, 0.0, 0.0]

        # Construct vects array
        self.vects = [[lx, 0.0, 0.0],
                      [xy, ly,  0.0],
                      [xz, yz,  lz]]
        self.origin = origin
    
    def set_hi_los(self, xlo, xhi, ylo, yhi, zlo, zhi, xy=0.0, xz=0.0, yz=0.0):
        """
        Set the box using LAMMPS box hi's, lo's and tilt factors.
        
        Parameters
        ----------
        xlo : float
            The LAMMPS xlo box lo term.
        xhi : float
            The LAMMPS xhi box hi term.
        ylo : float
            The LAMMPS ylo box lo term.
        yhi : float
            The LAMMPS yhi box hi term.
        zlo : float
            The LAMMPS zlo box lo term.
        zhi : float
            The LAMMPS zhi box hi term.
        xy : float, optional
            The LAMMPS box tilt factor in the xy direction.  Default value is 
            0.0.
        xz : float, optional
            The LAMMPS box tilt factor in the xz direction.  Default value is 
            0.0.
        yz : float, optional
            The LAMMPS box tilt factor in the yz direction.  Default value is 
            0.0.
        """
        
        # Convert to hi and lo term to lengths and origin
        lx = xhi - xlo
        ly = yhi - ylo
        lz = zhi - zlo
        origin = [xlo, ylo, zlo]
        
        # Call set_lengths
        self.set_lengths(lx=lx, ly=ly, lz=lz, xy=xy, xz=xz, yz=yz, origin=origin)
    
    def is_lammps_norm(self):
        """
        Tests if box is compatible with LAMMPS.
        Note: large box tilt factors not checked.  The LAMMPS command
        'box tilt large' may be needed to run LAMMPS.
        """
        return (self.__vects[0,1] == 0.0
            and self.__vects[0,2] == 0.0
            and self.__vects[1,2] == 0.0
            and self.__vects[0,0] > 0.0
            and self.__vects[1,1] > 0.0
            and self.__vects[2,2] > 0.0)

    def inside(self, pos, inclusive=True):
        
        # Retrieve the Box's planes
        planes = self.planes

        # Find all points below each plane, i.e. inside the box
        return ( planes[0].below(pos, inclusive=inclusive)
               & planes[1].below(pos, inclusive=inclusive)
               & planes[2].below(pos, inclusive=inclusive)
               & planes[3].below(pos, inclusive=inclusive)
               & planes[4].below(pos, inclusive=inclusive)
               & planes[5].below(pos, inclusive=inclusive))